#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.clock import Clock

from uvm.macros import *
from uvm import (UVMConfigDb, run_test)

from tb_env import tb_env
from apb.apb_if import apb_if
from testlib import *


@cocotb.test()
async def initial(dut):
    env = tb_env("env", None)
    vif = apb_if(dut)
    UVMConfigDb.set(env, "apb", "vif", vif)
    UVMConfigDb.set(None, "", "dut", dut)
    #UVMConfigDb.set(env, "apb", "vif", tb_top.apb0)
    cocotb.fork(Clock(vif.clk, 10, "NS").start())
    await run_test()



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.utils import simulator
from cocotb.clock import Clock
from uvm import (UVMConfigDb, run_test, UVMCoreService, sv,
    uvm_fatal, UVMUtils, uvm_hdl)
from apb.apb_if import apb_if

# Note: This has to be specified with PYTHONPATH
from tb_env import tb_env

@cocotb.test()
async def initial_begin(dut):
    cs_ = UVMCoreService.get()
    env = tb_env("env")

    print(str(dir(simulator)))
    print(str(dir(dut)))
    vif = apb_if(dut)
    
    uvm_hdl.set_dut(dut)
    #root = simulator.get_root_handle()
    #print(str(dir(root)))

    cocotb.fork(Clock(vif.clk, 10, "NS").start())
    svr = cs_.get_report_server()
    svr.set_max_quit_count(10)

    seq_name = []
    if sv.value_plusargs("UVM_SEQUENCE=%s", seq_name):
        seq = UVMUtils.create_type_by_name(seq_name[0], "tb")
        if seq is None:
            uvm_fatal("NO_SEQUENCE",
                "This env requires you to specify the sequence to run using UVM_SEQUENCE=<name>")
        env.seq = seq


    UVMConfigDb.set(env, "apb", "vif", vif)
    UVMConfigDb.set(None, "DUT_REF", "dut", dut)
    await run_test()



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Mentor Graphics Corporation
#//   Copyright 2010-2011 Synopsys, Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from uvm import (UVMComponent, uvm_component_utils, UVMTLMTime,
    UVMTLMBInitiatorSocket)

from uvm.macros import *
from apb_rw import apb_rw


class initiator(UVMComponent):
    #   uvm_tlm_b_initiator_socket#(apb_rw) sock


    def __init__(self, name="initiator", parent=None):
        super().__init__(name, parent)
        self.sock = UVMTLMBInitiatorSocket("sock", self)  # (apb_rw)("sock", self)

    #   //
    #   // Execute a simple read-modify-write
    #   //
    
    async def run_phase(self, phase):
        delay = UVMTLMTime()
        phase.raise_objection(self)
        for i in range(10):
            rw = apb_rw.type_id.create("rw",None,self.get_full_name())
            rw.kind = apb_rw.READ
            rw.addr = 0x0000FF00
            rw.data = i + 1

            await self.sock.b_transport(rw, delay)

            # Ok to reuse the same RW instance
            rw.kind = apb_rw.WRITE
            rw.data = ~rw.data

            await self.sock.b_transport(rw, delay)
        phase.drop_objection(self)


uvm_component_utils(initiator)



#//----------------------------------------------------------------------
#//   Copyright 2013 Cadence Design Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from .uvm_factory import UVMDefaultFactory


class UVMCoreService:
    """
    Class: UVMCoreService

    UVMCoreService provides a default implementation of the
    UVMCoreService API. It instantiates uvm_default_factory, uvm_default_report_server,
    uvm_root.
    """

    m_inst: 'UVMCoreService' = None

    def __init__(self):
        self.report_server = None
        self.factory = None
        self.tr_database = None  # uvm_tr_database
        self._visitor = None

    @classmethod
    def get(cls):
        """

        Returns:
            UVMCoreService: Core service singleton.

        """
        if UVMCoreService.m_inst is None:
            UVMCoreService.m_inst = UVMCoreService()
        return UVMCoreService.m_inst

    @classmethod
    def reset(cls):
        """Reset the state of coreservices. Used for unit testing"""
        UVMCoreService.m_inst = None
        from .uvm_registry import UVMComponentRegistry, UVMObjectRegistry
        UVMComponentRegistry.reset()
        UVMObjectRegistry.reset()

    def get_root(self):
        """
        Returns:
            UVMRoot: Singleton instance of UVMRoot.
        """
        from .uvm_root import UVMRoot
        return UVMRoot.m_uvm_get_root()

    def get_factory(self):
        """
        Function: get_factory

        Returns the currently enabled uvm factory.
        When no factory has been set before, instantiates a uvm_default_factory

        Returns:
            UVMFactory: Enabled UVM factory.
        """
        if self.factory is None:
            self.factory = UVMDefaultFactory()
        return self.factory

    def set_factory(self, factory):
        """
        Sets the current uvm factory.
        Please note: it is up to the user to preserve the contents of the original factory or
        delegate calls to the original factory

        Args:
            factory (UVMFactory):
        """
        self.factory = factory

    def get_default_tr_database(self):
        """
        Returns the current default record database

        If no default record database has been set before this method
        is called, returns an instance of `UVMTextTrDatabase`

        Returns:
            UVMTrDatabase: Default transaction database
        """
        if self.tr_database is None:
            #process p = process::self();
            p = None
            tx_db = None  # UVMTextTrDatabase
            s = ""
            if p is not None:
                s = p.get_randstate()

            from .uvm_tr_database import UVMTextTrDatabase
            tx_db = UVMTextTrDatabase("default_tr_database")
            self.tr_database = tx_db

            if p is not None:
                p.set_randstate(s)
        return self.tr_database


    def set_default_tr_database(self, db):
        """
        Function: set_default_tr_database
        Sets the current default record database to `db`

        Args:
            db (UVMTrDatabase): Default transaction database to use
        """
        self.tr_database = db

    def get_report_server(self):
        """
        Returns:
            UVMReportServer: Report server singleton instance.
        """
        if self.report_server is None:
            from .uvm_report_server import UVMReportServer
            self.report_server = UVMReportServer()
        return self.report_server

    def set_report_server(self, server):
        self.report_server = server

    #// Function: set_component_visitor
    #// sets the component visitor to ~v~
    #// (this visitor is being used for the traversal at end_of_elaboration_phase
    #// for instance for name checking)
    #virtual function void set_component_visitor(uvm_visitor#(uvm_component) v);
    #        _visitor=v;
    #endfunction


    #// Function: get_component_visitor
    #// retrieves the current component visitor
    #// if unset(or ~null~) returns a <uvm_component_name_check_visitor> instance
    #virtual function uvm_visitor#(uvm_component) get_component_visitor();
    #        if(_visitor==null) begin
    #                uvm_component_name_check_visitor v = new("name-check-visitor");
    #                _visitor=v;
    #        end
    #        return _visitor;
    #endfunction




from .uvm_pkg import m_uvm_string_queue_join

def uvm_typename(obj):
    name = "<NO CLASS NAME>"
    if hasattr(obj, '__class__'):
        name = obj.__class__
    else:
        name = obj
    return name

def UVM_STRING_QUEUE_STREAMING_PACK(q):
    return m_uvm_string_queue_join(q)



#//----------------------------------------------------------------------
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2010 Mentor Graphics Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#//------------------------
#// File: Register Defines
#//------------------------
#
#// Macro: `UVM_REG_ADDR_WIDTH
#//
#// Maximum address width in bits
#//
#// Default value is 64. Used to define the <uvm_reg_addr_t> type.
#//
#`ifndef UVM_REG_ADDR_WIDTH
# `define UVM_REG_ADDR_WIDTH 64
#`endif
#
#
#// Macro: `UVM_REG_DATA_WIDTH
#//
#// Maximum data width in bits
#//
#// Default value is 64. Used to define the <uvm_reg_data_t> type.
#//
#`ifndef UVM_REG_DATA_WIDTH
UVM_REG_DATA_WIDTH = 64
#`endif
#
#
#// Macro: `UVM_REG_BYTENABLE_WIDTH
#//
#// Maximum number of byte enable bits
#//
#// Default value is one per byte in <`UVM_REG_DATA_WIDTH>.
#// Used to define the <uvm_reg_byte_en_t> type.
#//
#`ifndef UVM_REG_BYTENABLE_WIDTH 
#  `define UVM_REG_BYTENABLE_WIDTH ((`UVM_REG_DATA_WIDTH-1)/8+1) 
#`endif
#
#
#// Macro: `UVM_REG_CVR_WIDTH
#//
#// Maximum number of bits in a <uvm_reg_cvr_t> coverage model set.
#//
#// Default value is 32.
#//
#`ifndef UVM_REG_CVR_WIDTH
# `define UVM_REG_CVR_WIDTH 32
#`endif
#



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

from ..base.uvm_component import UVMComponent
from ..tlm1.uvm_sqr_connections import UVMSeqItemPullPort
from ..tlm1.uvm_analysis_port import UVMAnalysisPort


class UVMDriver(UVMComponent):
    """
    The base class for drivers that initiate requests for new transactions via
    a uvm_seq_item_pull_port. The ports are typically connected to the exports of
    an appropriate sequencer component.

    This driver operates in pull mode. Its ports are typically connected to the
    corresponding exports in a pull sequencer as follows:

    .. code-block:: python
        driver.seq_item_port.connect(sequencer.seq_item_export);
        driver.rsp_port.connect(sequencer.rsp_export);

    The ~rsp_port~ needs connecting only if the driver will use it to write
    responses to the analysis export in the sequencer.

    :ivar UVMSeqItemPullPort seq_item_port: Derived driver classes should use
        this port to request items from the sequencer. They may also use it to
        send responses back.

    :ivar UVMAnalysisPort rst_port: This port provides an alternate way of
        sending responses back to the originating sequencer. Which port to
        use depends on which export the sequencer provides for connection.
    """



    #uvm_analysis_port #(RSP) rsp_port;

    # Function: new
    #
    # Creates and initializes an instance of this class using the normal
    # constructor arguments for <uvm_component>: ~name~ is the name of the
    # instance, and ~parent~ is the handle to the hierarchical parent, if any.
    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.seq_item_port    = UVMSeqItemPullPort("seq_item_port", self)
        self.rsp_port         = UVMAnalysisPort("rsp_port", self)
        self.seq_item_prod_if = self.seq_item_port
        self.req = None
        self.rsp = None

    type_name = "uvm_driver #(REQ,RSP)"

    def get_type_name(self):
        return UVMDriver.type_name




import unittest

from uvm.base.uvm_registry import UVMComponentRegistry, UVMObjectRegistry
from uvm.macros.uvm_object_defines import uvm_component_utils, uvm_object_utils


class TestUVMRegistry(unittest.TestCase):


    def test_object_registry(self):
        class ABC:
            def __init__(self, name):
                self.name = name
                self.value = 5
        uvm_object_utils(ABC)
        reg = UVMObjectRegistry(ABC, 'ABC')
        abc_obj = reg.create('obj_name')
        self.assertEqual(abc_obj.name, 'obj_name')
        self.assertEqual(isinstance(abc_obj, ABC), True)


    def test_component_registry(self):
        class ABC:
            def __init__(self, name, parent):
                self.name = name
                self.parent = parent
        uvm_component_utils(ABC)
        reg = UVMComponentRegistry(ABC, 'ABC')
        abc_comp = reg.create('inst_name', None)
        self.assertEqual(abc_comp.name, 'inst_name')
        self.assertEqual(abc_comp.parent, None)


if __name__ == '__main__':
    unittest.main()



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2008 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer

from ..uvm_reg_sequence import UVMRegSequence
from ..uvm_reg_model import (UVM_FRONTDOOR, UVM_IS_OK, UVM_NO_HIER,
    UVM_STATUS_NAMES)
from ...macros import (uvm_info, uvm_error, uvm_object_utils)
from ...base.uvm_resource_db import UVMResourceDb
from ...base.sv import sv
from ...base.uvm_object_globals import UVM_LOW
from ...base.uvm_globals import uvm_zero_delay

#//------------------------------------------------------------------------------
#// Title: Memory Walking-Ones Test Sequences
#//
#// This section defines sequences for applying a "walking-ones"
#// algorithm on one or more memories.
#//------------------------------------------------------------------------------


#//------------------------------------------------------------------------------
#// Class: UVMMemSingleWalkSeq
#//
#// Runs the walking-ones algorithm on the memory given by the <mem> property,
#// which must be assigned prior to starting this sequence.
#//
#// If bit-type resource named
#// "NO_REG_TESTS", "NO_MEM_TESTS", or "NO_MEM_WALK_TEST"
#// in the "REG::" namespace
#// matches the full name of the memory,
#// the memory is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.mem0.get_full_name()},
#//|                            "NO_MEM_TESTS", 1, this);
#//
#// The walking ones algorithm is performed for each map in which the memory
#// is defined.
#//
#//| for (k = 0 thru memsize-1)
#//|   write addr=k data=~k
#//|   if (k > 0) {
#//|     read addr=k-1, expect data=~(k-1)
#//|     write addr=k-1 data=k-1
#//|   if (k == last addr)
#//|     read addr=k, expect data=~k
#//
#//------------------------------------------------------------------------------


class UVMMemSingleWalkSeq(UVMRegSequence):  # (uvm_sequence #(uvm_reg_item))

    #   // Function: new
    #   //
    #   // Creates a new instance of the class with the given name.
    #
    def __init__(self, name="UVMMemWalkSeq"):
        super().__init__(name)
        #   // Variable: mem
        #   //
        #   // The memory to test; must be assigned prior to starting sequence.
        self.mem = None
    #   endfunction


    #   // Task: body
    #   //
    #   // Performs the walking-ones algorithm on each map of the memory
    #   // specified in <mem>.
    
    async def body(self):
        maps = []  # uvm_reg_map [$]
        n_bits = 0
        mem = self.mem

        if mem is None:
            uvm_error("UVMMemWalkSeq", "No memory specified to run sequence on")
            return

        # Memories with some attributes are not to be tested
        if (UVMResourceDb.get_by_name("REG::" + mem.get_full_name(),
            "NO_REG_TESTS", 0) is not None or
            UVMResourceDb.get_by_name("REG::" + mem.get_full_name(),
                "NO_MEM_TESTS", 0) is not None or
            UVMResourceDb.get_by_name("REG::" + mem.get_full_name(),
                "NO_MEM_WALK_TEST", 0) is not None):
            return

        n_bits = mem.get_n_bits()

        # Memories may be accessible from multiple physical interfaces (maps)
        mem.get_maps(maps)

        # Walk the memory via each map
        for j in range(len(maps)):
            status = 0
            val = 0
            exp = 0
            # v = 0

            # Only deal with RW memories
            if mem.get_access(maps[j]) != "RW":
                continue

            uvm_info("UVMMemWalkSeq", sv.sformatf("Walking memory %s (n_bits: %d) in map \"%s\"...",
                mem.get_full_name(), n_bits, maps[j].get_full_name()), UVM_LOW)

            # The walking process is, for address k:
            # - Write ~k
            # - Read k-1 and expect ~(k-1) if k > 0
            # - Write k-1 at k-1
            # - Read k and expect ~k if k == last address
            for k in range(mem.get_size()):
                status = []
                await mem.write(status, k, ~k, UVM_FRONTDOOR, maps[j], self)
                status = status[0]

                if status != UVM_IS_OK:
                    uvm_error("UVMMemWalkSeq", sv.sformatf(
                        "Status was %s when writing \"%s[%0d]\" through map \"%s\".",
                        status.name(), mem.get_full_name(), k, maps[j].get_full_name()))

                if k > 0:
                    status = []
                    val = []
                    await mem.read(status, k-1, val, UVM_FRONTDOOR, maps[j], self)
                    status = status[0]
                    if status != UVM_IS_OK:
                       uvm_error("UVMMemWalkSeq", sv.sformatf(
                           "Status was %s when reading \"%s[%0d]\" through map \"%s\".",
                           UVM_STATUS_NAMES[status], mem.get_full_name(), k, maps[j].get_full_name()))
                    else:
                        exp = ~(k-1) & ((1 << n_bits)-1)
                        val = val[0]
                        if val != exp:
                            uvm_error("UVMMemWalkSeq",
                                sv.sformatf("\"%s[%0d-1]\" read back as 'h%h instead of 'h%h.",
                                mem.get_full_name(), k, val, exp))

                    status = []
                    await mem.write(status, k-1, k-1, UVM_FRONTDOOR, maps[j], self)
                    status = status[0]
                    if status != UVM_IS_OK:
                        uvm_error("UVMMemWalkSeq", sv.sformatf(
                            "Status was %s when writing \"%s[%0d-1]\" through map \"%s\".",
                            UVM_STATUS_NAMES[status], mem.get_full_name(), k, maps[j].get_full_name()))

                if k == mem.get_size() - 1:
                    status = []
                    val = []
                    await mem.read(status, k, val, UVM_FRONTDOOR, maps[j], self)
                    status = status[0]
                    if status != UVM_IS_OK:
                        uvm_error("UVMMemWalkSeq", sv.sformatf(
                            "Status was %s when reading \"%s[%0d]\" through map \"%s\".",
                            UVM_STATUS_NAMES[status], mem.get_full_name(), k, maps[j].get_full_name()))
                    else:
                        exp = ~(k) & ((1<<n_bits)-1)
                        val = val[0]
                        if val != exp:
                            uvm_error("UVMMemWalkSeq", sv.sformatf("\"%s[%0d]\" read back as 'h%h instead of 'h%h.",
                                mem.get_full_name(), k, val, exp))


uvm_object_utils(UVMMemSingleWalkSeq)


#//------------------------------------------------------------------------------
#// Class: UVMMemWalkSeq
#//
#// Verifies the all memories in a block
#// by executing the <UVMMemSingleWalkSeq> sequence on
#// every memory within it.
#//
#// If bit-type resource named
#// "NO_REG_TESTS", "NO_MEM_TESTS", or "NO_MEM_WALK_TEST"
#// in the "REG::" namespace
#// matches the full name of the block,
#// the block is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.get_full_name(),".*"},
#//|                            "NO_MEM_TESTS", 1, this);
#//
#//------------------------------------------------------------------------------


class UVMMemWalkSeq(UVMRegSequence):  # (uvm_sequence #(uvm_reg_item))


    def __init__(self, name="UVMMemWalkSeq"):
        super().__init__(name)
        #   // Variable: model
        #   //
        #   // The block to be tested. Declared in the base class.
        #   //
        #   //| uvm_reg_block model;
        self.model = None
        #   // Variable: mem_seq
        #   //
        #   // The sequence used to test one memory
        #   //
        self.mem_seq = None


    #
    #   // Task: body
    #   //
    #   // Executes the mem walk sequence, one block at a time.
    #   // Do not call directly. Use seq.start() instead.
    #   //
    
    async def body(self):
        if self.model is None:
            uvm_error("UVMMemWalkSeq", "No register model specified to run sequence on")
            return

        uvm_info("STARTING_SEQ","\n\nStarting " + self.get_name() + " sequence...\n",UVM_LOW)

        self.mem_seq = UVMMemSingleWalkSeq.type_id.create("single_mem_walk_seq")

        await self.reset_blk(self.model)
        self.model.reset()
        await self.do_block(self.model)


    #   // Task: do_block
    #   //
    #   // Test all of the memories in a given ~block~
    #   //
    
    async def do_block(self, blk):
        mems = []  # uvm_mem[$]

        if (UVMResourceDb.get_by_name("REG::" + blk.get_full_name(),
            "NO_REG_TESTS", 0) is not None or
            UVMResourceDb.get_by_name("REG::" + blk.get_full_name(),
                "NO_MEM_TESTS", 0) is not None or
            UVMResourceDb.get_by_name("REG::" + blk.get_full_name(),
                "NO_MEM_ACCESS_TEST", 0) is not None):
            return

        # Iterate over all memories, checking accesses
        blk.get_memories(mems, UVM_NO_HIER)
        for i in range(len(mems)):
            # Memories with some attributes are not to be tested
            if (UVMResourceDb.get_by_name("REG::" + mems[i].get_full_name(),
                "NO_REG_TESTS", 0) is not None or
                UVMResourceDb.get_by_name("REG::" + mems[i].get_full_name(),
                    "NO_MEM_TESTS", 0) is not None or
                UVMResourceDb.get_by_name("REG::" + mems[i].get_full_name(),
                    "NO_MEM_WALK_TEST", 0) is not None):
                continue
            self.mem_seq.mem = mems[i]
            await self.mem_seq.start(None, self)

        blks = []  # uvm_reg_block [$]
        blk.get_blocks(blks)
        for i in range(len(blks)):
            await self.do_block(blks[i])


    #   // Task: reset_blk
    #   //
    #   // Reset the DUT that corresponds to the specified block abstraction class.
    #   //
    #   // Currently empty.
    #   // Will rollback the environment's phase to the ~reset~
    #   // phase once the new phasing is available.
    #   //
    #   // In the meantime, the DUT should be reset before executing this
    #   // test sequence or this method should be implemented
    #   // in an extension to reset the DUT.
    #   //
    
    async def reset_blk(self, blk):
        await uvm_zero_delay()


uvm_object_utils(UVMMemWalkSeq)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from uvm.base.uvm_object import UVMObject
from uvm.base.uvm_callback import UVMCallbackIter
from uvm.macros import (uvm_object_utils, uvm_fatal, uvm_do_obj_callbacks)
from uvm.reg.uvm_reg_model import UVM_NOT_OK
from uvm.base.uvm_globals import uvm_zero_delay
from .uvm_reg_cbs import UVMRegCbs

#//------------------------------------------------------------------------------
#// Class: UVMRegBackdoor
#//
#// Base class for user-defined back-door register and memory access.
#//
#// This class can be extended by users to provide user-specific back-door access
#// to registers and memories that are not implemented in pure SystemVerilog
#// or that are not accessible using the default DPI backdoor mechanism.
#//------------------------------------------------------------------------------


class UVMRegBackdoor(UVMObject):

    def __init__(self, name=""):
        """         
           Function: new
          
           Create an instance of self class
          
           Create an instance of the user-defined backdoor class
           for the specified register or memory
          
        Args:
            name: 
        """
        super().__init__(name)
        self.fname = ""
        self.lineno = 0

    async def do_pre_read(self, rw):
        """         

           Task: do_pre_read
          
           Execute the pre-read callbacks
          
           This method `must` be called as the first statement in
           a user extension of the <read()> method.
          
        Args:
            rw: 
        """
        await self.pre_read(rw)
        uvm_do_obj_callbacks(self, UVMRegCbs, 'pre_read', self, rw)

    async def do_post_read(self, rw):
        """         

           Task: do_post_read
          
           Execute the post-read callbacks
          
           This method `must` be called as the last statement in
           a user extension of the <read()> method.
          
        Args:
            rw: 
        """
        iter = UVMCallbackIter(self)
        # for(uvm_reg_cbs cb = iter.last(); cb is not None; cb=iter.prev())
        #     cb.decode(rw.value)
        cb = iter.first()
        while cb is not None:
            cb.decode(rw.value)
            cb = iter.next()
        uvm_do_obj_callbacks(self, UVMRegCbs,'post_read',self,rw)
        await self.post_read(rw)


    async def do_pre_write(self, rw):
        """         
           Task: do_pre_write
          
           Execute the pre-write callbacks
          
           This method `must` be called as the first statement in
           a user extension of the <write()> method.
          
        Args:
            rw: 
        """
        #uvm_callback_iter#(UVMRegBackdoor, uvm_reg_cbs) iter = new(self)
        iter = UVMCallbackIter(self)
        await self.pre_write(rw)
        uvm_do_obj_callbacks(self, UVMRegCbs,'pre_write',self,rw)
        #      for(uvm_reg_cbs cb = iter.first(); cb is not None; cb = iter.next())
        #         cb.encode(rw.value)
        cb = iter.first()
        while cb is not None:
            cb.decode(rw.value)
            cb = iter.next()


    async def do_post_write(self, rw):
        """         
           Task: do_post_write
          
           Execute the post-write callbacks
          
           This method `must` be called as the last statement in
           a user extension of the <write()> method.
        Args:
            rw: 
        """
        uvm_do_obj_callbacks(self, UVMRegCbs, 'post_write', self, rw)
        await self.post_write(rw)



    async def write(self, rw):  # task
        """         
           Task: write
          
           User-defined backdoor write operation.
          
           Call <do_pre_write()>.
           Deposit the specified value in the specified register HDL implementation.
           Call <do_post_write()>.
           Returns an indication of the success of the operation.
          
          extern def write(self,uvm_reg_item rw)
        Args:
            rw: 
        """
        uvm_fatal("RegModel", "UVMRegBackdoor::write() method has not been overloaded")


    async def read(self, rw):  # task:
        """         
           Task: read
          
           User-defined backdoor read operation.
          
           Overload self method only if the backdoor requires the use of task.
          
           Call <do_pre_read()>.
           Peek the current value of the specified HDL implementation.
           Call <do_post_read()>.
           Returns the current value and an indication of the success of
           the operation.
          
           By default, calls <read_func()>.
          
        Args:
            rw: 
        """
        await self.do_pre_read(rw)
        self.read_func(rw)
        await self.do_post_read(rw)


    def read_func(self, rw):
        """         
           Function: read_func
          
           User-defined backdoor read operation.
          
           Peek the current value in the HDL implementation.
           Returns the current value and an indication of the success of
           the operation.
          
          extern virtual def void read_func(self,uvm_reg_item rw):
        Args:
            rw: 
        """
        uvm_fatal("RegModel", "UVMRegBackdoor::read_func() method has not been overloaded")
        rw.status = UVM_NOT_OK
        #endfunction


    def is_auto_updated(self, field):
        """         
           Function: is_auto_updated
          
           Indicates if wait_for_change() method is implemented
          
           Implement to return TRUE if and only if
           <wait_for_change()> is implemented to watch for changes
           in the HDL implementation of the specified field
          
          extern virtual def bit is_auto_updated(self,uvm_reg_field field):
        Args:
            field: 
        Returns:
        """
        return False

    async def wait_for_change(self, element):  # task
        """         
           Task: wait_for_change
          
           Wait for a change in the value of the register or memory
           element in the DUT.
          
           When self method returns, the mirror value for the register
           corresponding to self instance of the backdoor class will be updated
           via a backdoor read operation.
          
        Args:
            element: 
        """
        uvm_fatal("RegModel", "UVMRegBackdoor::wait_for_change() method has not been overloaded")


    # tpoikela: Unsure if these are really needed anywhere?
    #   /*local*/ extern def void start_update_thread(self,uvm_object element):
    #   /*local*/ extern def void kill_update_thread(self,uvm_object element):
    #   /*local*/ extern def bit has_update_threads(self):


    async def pre_read(self, rw):
        """         
           Task: pre_read
          
           Called before user-defined backdoor register read.
          
           The registered callback methods are invoked after the invocation
           of self method.
          
        Args:
            rw: 
        """
        await uvm_zero_delay()


    async def post_read(self, rw):
        """         
           Task: post_read
          
           Called after user-defined backdoor register read.
          
           The registered callback methods are invoked before the invocation
           of self method.
        Args:
            rw: 
        """
        await uvm_zero_delay()


    async def pre_write(self, rw):
        """         
           Task: pre_write
          
           Called before user-defined backdoor register write.
          
           The registered callback methods are invoked after the invocation
           of self method.
          
           The written value, if modified, modifies the actual value that
           will be written.
        Args:
            rw: 
        """
        await uvm_zero_delay()


    async def post_write(self, rw):
        """         
           Task: post_write
          
           Called after user-defined backdoor register write.
          
           The registered callback methods are invoked before the invocation
           of self method.
        Args:
            rw: 
        """
        await uvm_zero_delay()


    #`ifdef UVM_USE_PROCESS_CONTAINER
    #   local process_container_c m_update_thread[uvm_object]
    #`else
    #   local process m_update_thread[uvm_object]
    #`endif


uvm_object_utils(UVMRegBackdoor)
#TODO   `uvm_register_cb(UVMRegBackdoor, uvm_reg_cbs)



#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#// start_update_thread
#
#def void UVMRegBackdoor::start_update_thread(self,uvm_object element):
#   uvm_reg rg
#   if (self.m_update_thread.exists(element)):
#      self.kill_update_thread(element)
#   end
#   if (!sv.cast(rg,element))
#     return; // only regs supported at self time
#
#   fork
#      begin
#         uvm_reg_field fields[$]
#
#`ifdef UVM_USE_PROCESS_CONTAINER
#         self.m_update_thread[element] = new(process::self())
#`else
#         self.m_update_thread[element] = process::self()
#`endif
#
#         rg.get_fields(fields)
#         while True:
#            uvm_status_e status
#            uvm_reg_data_t  val
#            uvm_reg_item r_item = new("bd_r_item")
#            r_item.element = rg
#            r_item.element_kind = UVM_REG
#            self.read(r_item)
#            val = r_item.value[0]
#            if (r_item.status != UVM_IS_OK):
#               `uvm_error("RegModel", sv.sformatf("Backdoor read of register '%s' failed.",
#                          rg.get_name()))
#            end
#            foreach (fields[i]):
#               if (self.is_auto_updated(fields[i])):
#                  r_item.value[0] = (val >> fields[i].get_lsb_pos()) &
#                                    ((1 << fields[i].get_n_bits())-1)
#                  fields[i].do_predict(r_item)
#                end
#            end
#            self.wait_for_change(element)
#         end
#      end
#   join_none
#endfunction
#
#
#// kill_update_thread
#
#def void UVMRegBackdoor::kill_update_thread(self,uvm_object element):
#   if (self.m_update_thread.exists(element)):
#
#`ifdef UVM_USE_PROCESS_CONTAINER
#      self.m_update_thread[element].p.kill()
#`else
#      self.m_update_thread[element].kill()
#`endif
#
#      self.m_update_thread.delete(element)
#   end
#endfunction
#
#
#// has_update_threads
#
#def bit UVMRegBackdoor::has_update_threads(self):
#   return self.m_update_thread.num() > 0
#endfunction



#
#-----------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#-----------------------------------------------------------------------------
"""
Title: Sequence Item Pull Ports

This section defines the port, export, and imp port classes for
communicating sequence items between `UVMSequencer` and
`UVMDriver`.
"""

from ..base.uvm_port_base import UVMPortBase
from ..macros.uvm_tlm_defines import *
from .uvm_tlm_imps import *




class UVMSeqItemPullPort():
    """
    Class: uvm_seq_item_pull_port #(REQ,RSP)
    
    UVM provides a port, export, and imp connector for use in sequencer-driver
    communication. All have standard port connector constructors, except that
    uvm_seq_item_pull_port's default min_size argument is 0; it can be left
    unconnected.
    """
    pass


UVMSeqItemPullPort = UVM_SEQ_PORT(UVMSeqItemPullPort,  # type: ignore
        UVM_SEQ_ITEM_PULL_MASK, "uvm_seq_item_pull_port")
UVM_SEQ_ITEM_PULL_IMP(UVMSeqItemPullPort, 'm_if')

#-----------------------------------------------------------------------------
#
# Class: uvm_seq_item_pull_export #(REQ,RSP)
#
# This export type is used in sequencer-driver communication. It has the
# standard constructor for exports.
#
#-----------------------------------------------------------------------------

class UVMSeqItemPullExport():
    pass
    #  extends uvm_port_base #(uvm_sqr_if_base #(REQ, RSP));
UVMSeqItemPullExport = UVM_EXPORT_COMMON(UVMSeqItemPullExport,  # type: ignore
        UVM_SEQ_ITEM_PULL_MASK, "uvm_seq_item_pull_export")
UVM_SEQ_ITEM_PULL_IMP(UVMSeqItemPullExport, 'm_if')

#-----------------------------------------------------------------------------
#
# Class: uvm_seq_item_pull_imp #(REQ,RSP,IMP)
#
# This imp type is used in sequencer-driver communication. It has the
# standard constructor for imp-type ports.
#
#-----------------------------------------------------------------------------

#class uvm_seq_item_pull_imp #(type REQ=int, type RSP=REQ, type IMP=int)
class UVMSeqItemPullImp():
    pass

UVMSeqItemPullImp = UVM_IMP_COMMON(UVMSeqItemPullImp,  # type: ignore
        UVM_SEQ_ITEM_PULL_MASK, "uvm_seq_item_pull_imp")
UVM_SEQ_ITEM_PULL_IMP(UVMSeqItemPullImp, 'm_imp')



#
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------


from ..base.uvm_object_globals import UVM_EXPORT, UVM_IMPLEMENTATION, UVM_PORT

# These IMP macros define implementations of the uvm_*_port, uvm_*_export,
# and uvm_*_imp ports.

#---------------------------------------------------------------
# Macros for implementations of UVM ports and exports
#
#/*
#`define UVM_BLOCKING_PUT_IMP(imp, TYPE, arg) \
#  task put (TYPE arg); \
#    if (m_imp_list.size()) == 0) begin \
#      uvm_report_error("Port Not Bound","Blocking put to unbound port will wait forever.", UVM_NONE);
#      @imp;
#    end
#    if (bcast_mode) begin \
#      if (m_imp_list.size()) > 1) \
#        fork
#          begin
#            foreach (m_imp_list[index]) \
#              fork \
#                automatic int i = index; \
#                begin m_imp_list[i].put(arg); end \
#              join_none \
#            wait fork; \
#          end \
#        join \
#      else \
#        m_imp_list[0].put(arg); \
#    end \
#    else  \
#      if (imp != null) \
#        imp.put(arg); \
#  endtask \
#
#`define UVM_NONBLOCKING_PUT_IMP(imp, TYPE, arg) \
#  function bit try_put(input TYPE arg); \
#    if (bcast_mode) begin \
#      if (!can_put()) \
#        return 0; \
#      foreach (m_imp_list[index]) \
#        void'(m_imp_list[index].try_put(arg)); \
#      return 1; \
#    end  \
#    if (imp != null) \
#      return imp.try_put(arg)); \
#    return 0; \
#  endfunction \
#  \
#  function bit can_put(); \
#    if (bcast_mode) begin \
#      if (m_imp_list.size()) begin \
#        foreach (m_imp_list[index]) begin \
#          if (!m_imp_list[index].can_put() \
#            return 0; \
#        end \
#        return 1; \
#      end \
#      return 0; \
#    end \
#    if (imp != null) \
#      return imp.can_put(); \
#    return 0; \
#  endfunction
#
#*/

#-----------------------------------------------------------------------
# TLM imp implementations
#
#`define UVM_BLOCKING_PUT_IMP(imp, TYPE, arg) \
def UVM_BLOCKING_PUT_IMP(imp, T):
    
    async def put(self, t):
        if not hasattr(self, imp):
            raise Exception('Tried calling put() for attr {}'.format(imp))
        port_imp = getattr(self, imp)
        await port_imp.put(t)
    #endtask
    setattr(T, 'put', put)

#`define UVM_NONBLOCKING_PUT_IMP(imp, TYPE, arg) \
#  function bit try_put (TYPE arg); \
#    return imp.try_put(arg); \
#  endfunction \
#  function bit can_put(); \
#    return imp.can_put(); \
#  endfunction
def UVM_NONBLOCKING_PUT_IMP(imp, T):
    def try_put(self, arg):
        return getattr(self, imp).try_put(arg)
    setattr(T, 'try_put', try_put)

    def can_put(self):
        return getattr(self, imp).can_put()
    setattr(T, 'can_put', can_put)

#`define UVM_BLOCKING_GET_IMP(imp, TYPE, arg) \
#  task get (output TYPE arg); \
#    imp.get(arg); \
#  endtask
def UVM_BLOCKING_GET_IMP(imp, T):
    
    async def get(self, arg):
        await getattr(self, imp).get(arg)
    setattr(T, 'get', get)

#`define UVM_NONBLOCKING_GET_IMP(imp, TYPE, arg) \
#  function bit try_get (output TYPE arg); \
#    return imp.try_get(arg); \
#  endfunction \
#  function bit can_get(); \
#    return imp.can_get(); \
#  endfunction
def UVM_NONBLOCKING_GET_IMP(imp, T):
    def try_get(self, arg):
        return getattr(self, imp).try_get(arg)
    setattr(T, 'try_get', try_get)

    def can_get(self):
        return getattr(self, imp).can_get()
    setattr(T, 'can_get', can_get)

#`define UVM_BLOCKING_PEEK_IMP(imp, TYPE, arg) \
#  task peek (output TYPE arg); \
#    imp.peek(arg); \
def UVM_BLOCKING_PEEK_IMP(imp, T):
    
    async def peek(self, arg):
        await getattr(self, imp).peek(arg)
    setattr(T, 'peek', peek)

#`define UVM_NONBLOCKING_PEEK_IMP(imp, TYPE, arg) \
#  function bit try_peek (output TYPE arg); \
#    return imp.try_peek(arg); \
#  function bit can_peek(); \
#    return imp.can_peek(); \

def UVM_NONBLOCKING_PEEK_IMP(imp, T):
    def try_peek(self, arg):
        return getattr(self, imp).try_peek(arg)
    setattr(T, 'try_peek', try_peek)

    def can_peek(self):
        return getattr(self, imp).can_peek()
    setattr(T, 'can_peek', can_peek)

#`define UVM_BLOCKING_TRANSPORT_IMP(imp, REQ, RSP, req_arg, rsp_arg) \
#  task transport (REQ req_arg, output RSP rsp_arg); \
#    imp.transport(req_arg, rsp_arg); \
def UVM_BLOCKING_TRANSPORT_IMP(imp, T):
    
    async def transport (self, req_arg, rsp_arg):
        await getattr(self, imp).transport(req_arg, rsp_arg)
    setattr(T, 'transport', transport)

#`define UVM_NONBLOCKING_TRANSPORT_IMP(imp, REQ, RSP, req_arg, rsp_arg) \
#  function bit nb_transport (REQ req_arg, output RSP rsp_arg); \
#    return imp.nb_transport(req_arg, rsp_arg); \
def UVM_NONBLOCKING_TRANSPORT_IMP(imp, T):
    def nb_transport(self, req_arg, rsp_arg):
        return getattr(self, imp).nb_transport(req_arg, rsp_arg)
    setattr(T, 'nb_transport', nb_transport)

def UVM_PUT_IMP(imp, T):
    UVM_BLOCKING_PUT_IMP(imp, T)
    UVM_NONBLOCKING_PUT_IMP(imp, T)

def UVM_GET_IMP(imp, T):
    UVM_BLOCKING_GET_IMP(imp, T)
    UVM_NONBLOCKING_GET_IMP(imp, T)

#`define UVM_PEEK_IMP(imp, TYPE, arg) \
#  `UVM_BLOCKING_PEEK_IMP(imp, TYPE, arg) \
#  `UVM_NONBLOCKING_PEEK_IMP(imp, TYPE, arg)
def UVM_PEEK_IMP(imp, T):
    UVM_BLOCKING_PEEK_IMP(imp, T)
    UVM_NONBLOCKING_PEEK_IMP(imp, T)

#`define UVM_BLOCKING_GET_PEEK_IMP(imp, TYPE, arg) \
#  `UVM_BLOCKING_GET_IMP(imp, TYPE, arg) \
#  `UVM_BLOCKING_PEEK_IMP(imp, TYPE, arg)
def UVM_BLOCKING_GET_PEEK_IMP(imp, T):
    UVM_BLOCKING_GET_IMP(imp, T)
    UVM_BLOCKING_PEEK_IMP(imp, T)

#`define UVM_NONBLOCKING_GET_PEEK_IMP(imp, TYPE, arg) \
#  `UVM_NONBLOCKING_GET_IMP(imp, TYPE, arg) \
#  `UVM_NONBLOCKING_PEEK_IMP(imp, TYPE, arg)
def UVM_NONBLOCKING_GET_PEEK_IMP(imp, T):
    UVM_NONBLOCKING_GET_IMP(imp, T)
    UVM_NONBLOCKING_PEEK_IMP(imp, T)

#`define UVM_GET_PEEK_IMP(imp, TYPE, arg) \
#  `UVM_BLOCKING_GET_PEEK_IMP(imp, TYPE, arg) \
#  `UVM_NONBLOCKING_GET_PEEK_IMP(imp, TYPE, arg)
def UVM_GET_PEEK_IMP(imp, T):
    UVM_BLOCKING_GET_PEEK_IMP(imp, T)
    UVM_NONBLOCKING_GET_PEEK_IMP(imp, T)

#`define UVM_TRANSPORT_IMP(imp, REQ, RSP, req_arg, rsp_arg) \
#  `UVM_BLOCKING_TRANSPORT_IMP(imp, REQ, RSP, req_arg, rsp_arg) \
#  `UVM_NONBLOCKING_TRANSPORT_IMP(imp, REQ, RSP, req_arg, rsp_arg)
def UVM_TRANSPORT_IMP(imp, T):
    UVM_BLOCKING_TRANSPORT_IMP(imp, T)
    UVM_NONBLOCKING_TRANSPORT_IMP(imp, T)

def UVM_TLM_GET_TYPE_NAME(T):
    Ts = T.__name__

    def get_type_name(self):
        return Ts
    setattr(T, 'get_type_name', get_type_name)
#

#`define UVM_PORT_COMMON(MASK,TYPE_NAME) \
#  function new (string name, uvm_component parent, \
#                int min_size=1, int max_size=1); \
#    super.new (name, parent, UVM_PORT, min_size, max_size); \
#    m_if_mask = MASK; \
#  endfunction \
#  UVM_TLM_GET_TYPE_NAME(TYPE_NAME)
def UVM_PORT_COMMON(T, MASK, TYPE_NAME):
    from ..base.uvm_port_base import UVMPortBase
    def __init__(self, name, parent, min_size=1, max_size=1):
        UVMPortBase.__init__(self, name, parent, UVM_PORT, min_size, max_size)
        self.m_if_mask = MASK
    setattr(T, '__init__', __init__)
    UVM_TLM_GET_TYPE_NAME(T)
    return add_base_class(T, UVMPortBase)

#
#`define UVM_SEQ_PORT(MASK,TYPE_NAME) \
#  function new (string name, uvm_component parent, \
#                int min_size=0, int max_size=1); \
#    super.new (name, parent, UVM_PORT, min_size, max_size); \
#    m_if_mask = MASK; \
#  endfunction \
#  UVM_TLM_GET_TYPE_NAME(TYPE_NAME)
def UVM_SEQ_PORT(T, MASK, TYPE_NAME):
    from ..base.uvm_port_base import UVMPortBase
    def __init__(self, name, parent, min_size=0, max_size=1):
        UVMPortBase.__init__(self, name, parent, UVM_PORT, min_size, max_size)
        self.m_if_mask = MASK
    setattr(T, '__init__', __init__)
    UVM_TLM_GET_TYPE_NAME(T)
    return add_base_class(T, UVMPortBase)

#`define UVM_EXPORT_COMMON(MASK,TYPE_NAME) \
#  function new (string name, uvm_component parent, \
#                int min_size=1, int max_size=1); \
#    super.new (name, parent, UVM_EXPORT, min_size, max_size); \
#    m_if_mask = MASK; \
def UVM_EXPORT_COMMON(T,MASK,TYPE_NAME):
    from ..base.uvm_port_base import UVMPortBase
    def __init__(self, name, parent, min_size=1, max_size=1):
        UVMPortBase.__init__(self, name, parent, UVM_EXPORT, min_size, max_size)
        self.m_if_mask = MASK
    setattr(T, '__init__', __init__)
    UVM_TLM_GET_TYPE_NAME(T)
    return add_base_class(T, UVMPortBase)

def UVM_EXPORT_PORT(T, MASK, TYPE_NAME):
    from ..base.uvm_port_base import UVMPortBase
    def __init__(self, name, parent, min_size=1, max_size=1):
        UVMPortBase.__init__(self, name, parent, UVM_EXPORT, min_size, max_size)
        self.m_if_mask = MASK
    setattr(T, '__init__', __init__)
    UVM_TLM_GET_TYPE_NAME(T)
    return add_base_class(T, UVMPortBase)

def UVM_IMP_COMMON(T,MASK,TYPE_NAME):
    from ..base.uvm_port_base import UVMPortBase
    def __init__(self, name, imp):
        UVMPortBase.__init__(self, name, imp, UVM_IMPLEMENTATION, 1, 1)
        self.m_imp = imp
        self.m_if_mask = MASK
    UVM_TLM_GET_TYPE_NAME(T)
    TT = add_base_class(T, UVMPortBase)
    setattr(TT, '__init__', __init__)
    return TT

def add_base_class(T, new_base, args={}):
    return type(T.__name__, (T, new_base,), args)

#`define UVM_MS_IMP_COMMON(MASK,TYPE_NAME) \
#  local this_req_type m_req_imp; \
#  local this_rsp_type m_rsp_imp; \
#  function new (string name, this_imp_type imp, \
#                this_req_type req_imp = null, this_rsp_type rsp_imp = null); \
#    super.new (name, imp, UVM_IMPLEMENTATION, 1, 1); \
#    if(req_imp==null) $cast(req_imp, imp); \
#    if(rsp_imp==null) $cast(rsp_imp, imp); \
#    m_req_imp = req_imp; \
#    m_rsp_imp = rsp_imp; \
#    m_if_mask = MASK; \
#  endfunction  \
#  UVM_TLM_GET_TYPE_NAME(TYPE_NAME)
def UVM_MS_IMP_COMMON(T, MASK, TYPE_NAME):
    from ..base.uvm_port_base import UVMPortBase
    def __init__(self, name, imp, req_imp=None, rsp_imp=None):
        UVMPortBase.__init__(self, name, imp, UVM_IMPLEMENTATION, 1, 1)
        if req_imp is None:
            req_imp = imp
        if rsp_imp is None:
            rsp_imp = imp
        self.m_req_imp = req_imp
        self.m_rsp_imp = rsp_imp
        self.m_if_mask = MASK
    setattr(T, '__init__', __init__)
    UVM_TLM_GET_TYPE_NAME(T)
    return add_base_class(T, UVMPortBase)





import unittest
from uvm.base.uvm_pool import UVMPool, UVMEventPool


class TestUVMPool(unittest.TestCase):

    def test_global_pool(self):
        gpool = UVMPool.get_global_pool()
        gpool.add('abc', 879)
        g_val = UVMPool.get_global('abc')
        self.assertEqual(g_val, 879)

    def test_pool(self):
        pool = UVMPool()
        pool.add('xxx', 123)
        self.assertEqual(pool.num(), 1)
        self.assertEqual(pool.get('xxx'), 123)
        first_key = pool.first()
        last_key = pool.last()
        self.assertEqual(first_key, 'xxx')
        self.assertEqual(first_key, last_key)

    def test_pool_iter(self):
        pool = UVMPool()
        pool.add('a', 1)
        pool.add('b', 2)
        pool.add('c', 3)
        self.assertEqual(pool.num(), 3)
        self.assertEqual(pool.has_prev(), False)
        self.assertEqual(pool.has_next(), True)
        self.assertEqual(pool.last(), 'c')
        self.assertEqual(pool.first(), 'a')
        self.assertEqual(pool.last(), 'c')
        self.assertEqual(pool.prev(), 'b')
        self.assertEqual(pool.prev(), 'a')
        self.assertEqual(pool.next(), 'b')
        self.assertEqual(pool.next(), 'c')

    def test_len_and_in(self):
        pool = UVMPool()
        self.assertNotIn('xxx', pool)
        self.assertEqual(len(pool), 0)
        pool.add('xxx', 1234)
        self.assertIn('xxx', pool)
        self.assertEqual(len(pool), 1)
        empty = UVMPool()
        self.assertNotIn('xxx', empty)

    def test_while_has_next(self):
        pool = UVMPool()
        pool.add(1, 2)
        pool.add(4, 8)
        pool.add(8, 16)
        while (pool.has_next()):
            val = pool.next()
            self.assertTrue(pool.exists(val))
            self.assertIn(pool[val], [2, 8, 16])

    def test_for_loop(self):
        pool = UVMPool()
        pool.add("z", 0)
        pool.add("a", 4)
        pool.add("b", 8)
        pool.add("c", 16)
        for key in pool:
            val = pool[key]
            self.assertTrue(val in [0, 4, 8, 16])
            self.assertTrue(key in ["z", "a", "b", "c"])


class TestUVMEventPool(unittest.TestCase):

    def test_event_pool(self):
        event_pool = UVMEventPool('event_pool')
        evt_start = event_pool.get('start')
        evt_end = event_pool.get('end')
        self.assertEqual(evt_start.get_name(), 'start')
        self.assertEqual(evt_end.get_name(), 'end')
        evt_end2 = event_pool.get('end')
        self.assertIs(evt_end, evt_end2)
        self.assertNotEqual(evt_end, evt_start)


if __name__ == '__main__':
    unittest.main()



#//
#// -------------------------------------------------------------
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2010-2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

# Note: You can also import everything with 'from uvm import *'
from uvm.comps import UVMEnv
from uvm.macros import uvm_component_utils
from uvm.reg import UVMRegPredictor

from reg_model import reg_block_B
from apb.apb_agent import apb_agent
from apb.apb_rw import reg2apb_adapter


class tb_env(UVMEnv):

    #
    #   reg_block_B   regmodel
    #   apb_agent     apb
    #   uvm_reg_predictor#(apb_rw) predict

    def __init__(self, name="tb_env", parent=None):
        super().__init__(name, parent)
        self.regmodel = None
        self.apb = None
        self.predict = None


    def build_phase(self, phase):
        if self.regmodel is None:
            self.regmodel = reg_block_B.type_id.create("regmodel")
            self.regmodel.build()
            self.regmodel.lock_model()

        self.apb = apb_agent.type_id.create("apb", self)
        self.predict = UVMRegPredictor.type_id.create("predict", self)


    def connect_phase(self, phase):
        if self.regmodel.get_parent() is None:
            self.apb_adapter  = reg2apb_adapter("apb_adapter")
            self.regmodel.default_map.set_sequencer(self.apb.sqr, self.apb_adapter)
            self.regmodel.default_map.set_auto_predict(0)

            self.apb.mon.ap.connect(self.predict.bus_in)

            self.predict.map = self.regmodel.default_map
            self.predict.adapter = self.apb_adapter


uvm_component_utils(tb_env)



#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#// 


class vip_tx_if():

    def __init__(self, clk, Tx):
        self.clk = clk
        self.Tx = Tx

class vip_rx_if():

    def __init__(self, clk, Rx):
        self.clk = clk
        self.Rx = Rx

class vip_if():

    def __init__(self, clk, Tx, Rx):
        self.clk = clk
        self.Tx = Tx
        self.Rx = Rx

        self.tx = vip_tx_if(self.clk, self.Tx)
        self.rx = vip_rx_if(self.clk, self.Rx)
        self.tx_mon = vip_rx_if(self.clk, self.Tx)



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------


from cocotb.triggers import Event

from .uvm_imps import UVMGetPeekImp, UVMPutImp
from .uvm_analysis_port import UVMAnalysisPort
from ..base.uvm_component import UVMComponent
from ..base.uvm_object_globals import UVM_NONE
from ..base.uvm_globals import uvm_report_error

UVM_TLM_FIFO_TASK_ERROR = "fifo channel task not implemented"
UVM_TLM_FIFO_FUNCTION_ERROR = "fifo channel function not implemented"


class UVMTLMEvent:
    def __init__(self):
        self.event = Event()


class UVMTLMFIFOBase(UVMComponent):
    """
    This class is the base for `UVMTLMFIFO`. It defines the TLM exports
    through which all transaction-based FIFO operations occur. It also defines
    default implementations for each interface method provided by these exports.

    The interface methods provided by the `put_export` and the `get_peek_export`
    are defined and described by `UVMTLMIfBase`.  See the TLM Overview
    section for a general discussion of TLM interface definition and usage.

    :ivar UVMPutImp put_export:

     The `put_export` provides both the blocking and non-blocking put interface
     methods to any attached port:

     Example:

     .. code-block:: python

        task put (input T t)
        function bit can_put ()
        function bit try_put (input T t)

    Any `put` port variant can connect and send transactions to the FIFO via this
    export, provided the transaction types match. See `UVMTLMIfBase`
    for more information on each of the above interface methods.

    :ivar UVMGetPeekImp get_peek_export:

    The `get_peek_export` provides all the blocking and non-blocking get and peek
    interface methods:

    .. code-block:: python
        task get (output T t)
        function bit can_get ()
        function bit try_get (output T t)
        task peek (output T t)
        function bit can_peek ()
        function bit try_peek (output T t)

    Any `get` or `peek` port variant can connect to and retrieve transactions from
    the FIFO via this export, provided the transaction types match. See
    <uvm_tlm_if_base #(T1,T2)> for more information on each of the above interface
    methods.

    :ivar UVMAnalysisPort put_ap:

    Transactions passed via `put` or `try_put` (via any port connected to the
    <put_export>) are sent out this port via its `write` method.

    .. code-block:: python
        def write (T t)

    All connected analysis exports and imps will receive put transactions.
    See <uvm_tlm_if_base #(T1,T2)> for more information on the `write` interface
    method.

    :ivar UVMAnalysisPort get_ap:

    Transactions passed via `get`, `try_get`, `peek`, or `try_peek` (via any
    port connected to the <get_peek_export>) are sent out this port via its
    `write` method.

    .. code-block:: python
        def write (T t)

    All connected analysis exports and imps will receive get transactions.
    See <uvm_tlm_if_base #(T1,T2)> for more information on the `write` method.

    The following are aliases to the above put_export.

    :ivar UVMPutImp blocking_put_export:
    :ivar UVMPutImp nonblocking_put_export:

    The following are all aliased to the above get_peek_export, which provides
    the superset of these interfaces.

    :ivar `UVMGetPeekImp` blocking_get_export:
    :ivar `UVMGetPeekImp` nonblocking_get_export:
    :ivar `UVMGetPeekImp` get_export:
    :ivar `UVMGetPeekImp` blocking_peek_export:
    :ivar `UVMGetPeekImp` nonblocking_peek_export:
    :ivar `UVMGetPeekImp` peek_export:
    :ivar `UVMGetPeekImp` blocking_get_peek_export:
    :ivar `UVMGetPeekImp` nonblocking_get_peek_export:
    """

    def __init__(self, name, parent=None):
        """
          The `name` and `parent` are the normal `UVMComponent` constructor arguments.
          The `parent` should be `None` if the `UVMTLMFIFO` is going to be used in a
          statically elaborated construct (e.g., a module). The `size` indicates the
          maximum size of the FIFO. A value of zero indicates no upper bound.

        Args:
            name (str): Name of the component
            parent (UVMComponent): Parent component.
        """
        UVMComponent.__init__(self, name, parent)
        self.put_export = UVMPutImp("put_export", self)
        self.blocking_put_export = self.put_export
        self.nonblocking_put_export = self.put_export

        get_peek_export = UVMGetPeekImp("get_peek_export", self)
        self.get_peek_export = get_peek_export
        self.blocking_get_peek_export = get_peek_export
        self.nonblocking_get_peek_export = get_peek_export
        self.blocking_get_export = get_peek_export
        self.nonblocking_get_export = get_peek_export
        self.get_export = get_peek_export
        self.blocking_peek_export = get_peek_export
        self.nonblocking_peek_export = get_peek_export
        self.peek_export = get_peek_export
        self.put_ap = UVMAnalysisPort("put_ap", self)
        self.get_ap = UVMAnalysisPort("get_ap", self)

    def build_phase(self, phase):
        """
        Args:
            phase (UVMPhase):
        """
        self.build()  # for backward compat, won't cause auto-config


    def flush(self):
        uvm_report_error("flush", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)

    def size(self):
        uvm_report_error("size", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0


    async def put(self, t):
        uvm_report_error("put", UVM_TLM_FIFO_TASK_ERROR, UVM_NONE)


    async def get(self,t):
        uvm_report_error("get", UVM_TLM_FIFO_TASK_ERROR, UVM_NONE)


    async def peek(self, t):
        uvm_report_error("peek", UVM_TLM_FIFO_TASK_ERROR, UVM_NONE)

    def try_put(self, t):
        uvm_report_error("try_put", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def try_get(self, t):
        uvm_report_error("try_get", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def try_peek(self, t):
        uvm_report_error("try_peek", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def can_put(self):
        uvm_report_error("can_put", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def can_get(self):
        uvm_report_error("can_get", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def can_peek(self):
        uvm_report_error("can_peek", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def ok_to_put(self):
        uvm_report_error("ok_to_put", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return None

    def ok_to_get(self):
        uvm_report_error("ok_to_get", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return None

    def ok_to_peek(self):
        uvm_report_error("ok_to_peek", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return None

    def is_empty(self):
        uvm_report_error("is_empty", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0

    def is_full(self):
        uvm_report_error("is_full", UVM_TLM_FIFO_FUNCTION_ERROR)
        return 0

    def used(self):
        uvm_report_error("used", UVM_TLM_FIFO_FUNCTION_ERROR, UVM_NONE)
        return 0




import unittest

from uvm.reg.uvm_reg_predictor import UVMRegPredictor, UVMPredictS

from uvm.uvm_unit import (create_reg, create_reg_block, TestPacket,
    TestRegAdapter)


class TestUVMRegPredictor(unittest.TestCase):

    def test_create_predictor(self):
        predict = UVMRegPredictor("predictor_123", None)
        self.assertEqual(predict.get_name(), "predictor_123")

    def test_predict(self):
        predict = UVMRegPredictor("predictor_345", None)
        predict.adapter = TestRegAdapter()
        rg = create_reg_block('my_block')
        rg.lock_model()
        predict.map = rg.default_map
        predict.bus_in.write(TestPacket(123, 0x0))

    def test_check_phase(self):
        predict = UVMRegPredictor("predictor_567", None)
        predict.check_phase(phase=None)
        predict.m_pending[create_reg('test_reg')] = UVMPredictS()
        predict.check_phase(phase=None)


if __name__ == '__main__':
    unittest.main()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
#from cocotb.clock import Clock
from uvm.base import run_test, UVMDebug
from uvm.base.uvm_phase import UVMPhase

from test_lib import *
from ubus_if import ubus_if

UBUS_ADDR_WIDTH = 16

#UVMDebug.DEBUG = True
#UVMPhase.m_phase_trace = True



async def initial_run_test(dut, vif):
    from uvm.base import UVMCoreService
    cs_ = UVMCoreService.get()
    UVMConfigDb.set(None, "*", "vif", vif)
    await run_test()



async def initial_reset(vif):
    await Timer(5, "NS")
    vif.ubus_reset <= 1
    vif.ubus_clock <= 1
    vif.ubus_wait <= 0
    await Timer(51, "NS")
    vif.ubus_reset <= 0



async def always_clk(dut, ncycles):
    dut.ubus_clock <= 0
    n = 0
    print("EEE starting always_clk")
    while n < 2*ncycles:
        n += 1
        await Timer(5, "NS")
        next_val = not dut.ubus_clock.value
        dut.ubus_clock <= int(next_val)

#module ubus_tb_top;
@cocotb.test()
async def module_ubus_tb(dut):

    vif = ubus_if(dut)

    proc_run_test = cocotb.fork(initial_run_test(dut, vif))
    proc_reset = cocotb.fork(initial_reset(dut))
    proc_clk = cocotb.fork(always_clk(dut, 100))
    proc_vif = cocotb.fork(vif.start())

    await Timer(999, "NS")
    #await [proc_run_test, proc_clk.join()]
    await sv.fork_join([proc_run_test, proc_clk])



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------

from ..base.uvm_object import UVMObject
from ..base.uvm_pool import UVMPool, UVMObjectStringPool
from ..base.uvm_queue import UVMQueue
from ..base.uvm_globals import (uvm_check_output_args, uvm_zero_delay, uvm_report_enabled,
    uvm_report_info)
from ..base.uvm_object_globals import UVM_DEBUG, UVM_HIGH, UVM_INFO
from ..base.sv import sv
from ..macros.uvm_message_defines import uvm_error, uvm_info, uvm_warning, uvm_fatal
from .uvm_reg_model import (UVM_BACKDOOR, UVM_BURST_READ, UVM_BURST_WRITE, UVM_DEFAULT_PATH,
                            UVM_FRONTDOOR, UVM_IS_OK, UVM_MEM, UVM_NOT_OK, UVM_NO_COVERAGE,
                            UVM_READ, UVM_WRITE, uvm_hdl_concat2string, uvm_hdl_path_concat,
                            uvm_reg_cvr_rsrc_db)
from .uvm_reg_item import UVMRegItem
from .uvm_reg_map import UVMRegMap
from .uvm_mem_mam import UVMMemMam, UVMMemMamCfg
from .uvm_reg_cbs import UVMMemCbIter

from ..dpi.uvm_hdl import uvm_hdl

ERR_MEM_BD_READ = ("Backdoor read of register %s with multiple HDL copies: "
    + "values are not the same: %0h at path '%s', and %0h at path '%s'. Returning first value.")


#   typedef enum {UNKNOWNS, ZEROES, ONES, ADDRESS, VALUE, INCR, DECR} init_e
UNKNOWNS = 0
ZEROES = 1
ONES = 2
ADDRESS = 3
VALUE = 4
INCR = 5
DECR = 6


class UVMMem(UVMObject):
    """
    Memory abstraction base class

    A memory is a collection of contiguous locations.
    A memory may be accessible via more than one address map.
    
    Unlike registers, memories are not mirrored because of the potentially
    large data space: tests that walk the entire memory space would negate
    any benefit from sparse memory modelling techniques.
    Rather than relying on a mirror, it is recommended that
    backdoor access be used instead.
    """



    #   local uvm_object_string_pool
    #               #(uvm_queue #(uvm_hdl_path_concat)) m_hdl_paths_pool
    #

    m_max_size = 0

    #   //----------------------
    #   // Group: Initialization
    #   //----------------------

    #   // Function: new
    #   //
    #   // Create a new instance and type-specific configuration
    #   //
    #   // Creates an instance of a memory abstraction class with the specified
    #   // name.
    #   //
    #   // ~size~ specifies the total number of memory locations.
    #   // ~n_bits~ specifies the total number of bits in each memory location.
    #   // ~access~ specifies the access policy of this memory and may be
    #   // one of "RW for RAMs and "RO" for ROMs.
    #   //
    #   // ~has_coverage~ specifies which functional coverage models are present in
    #   // the extension of the register abstraction class.
    #   // Multiple functional coverage models may be specified by adding their
    #   // symbolic names, as defined by the <uvm_coverage_model_e> type.
    #   //
    #   extern function new (string           name,
    #                        longint unsigned size,
    #                        int unsigned     n_bits,
    #                        string           access = "RW",
    #                        int              has_coverage = UVM_NO_COVERAGE)
    def __init__(self, name, size, n_bits, access="RW",
            has_coverage=UVM_NO_COVERAGE):
        UVMObject.__init__(self, name)
        self.m_locked = 0
        if n_bits == 0:
            uvm_error("RegModel", "Memory '" + self.get_full_name() + "' cannot have 0 bits")
            n_bits = 1

        self.m_fname = ""
        self.m_parent = None
        self.m_size      = size
        self.m_n_bits    = n_bits
        self.m_backdoor  = None
        self.m_access    = access.upper()
        self.m_has_cover = has_coverage
        self.m_hdl_paths_pool = UVMObjectStringPool("hdl_paths", UVMQueue)
        self.m_read_in_progress = False
        self.m_write_in_progress = False
        self.m_is_powered_down = False
        self.m_cover_on = 0

        self.m_vregs = UVMPool()  # [uvm_vreg]
        self.m_maps = UVMPool()  # bit[uvm_reg_map]
        if (n_bits > UVMMem.m_max_size):
            UVMMem.m_max_size = n_bits

        #   // variable: mam
        #   //
        #   // Memory allocation manager
        #   //
        #   // Memory allocation manager for the memory corresponding to this
        #   // abstraction class instance.
        #   // Can be used to allocate regions of consecutive addresses of
        #   // specific sizes, such as DMA buffers,
        #   // or to locate virtual register array.
        #   //
        self.mam = None  # uvm_mem_mam


    #   // Function: configure
    #   //
    #   // Instance-specific configuration
    #   //
    #   // Specify the parent block of this memory.
    #   //
    #   // If this memory is implemented in a single HDL variable,
    #   // its name is specified as the ~hdl_path~.
    #   // Otherwise, if the memory is implemented as a concatenation
    #   // of variables (usually one per bank), then the HDL path
    #   // must be specified using the <add_hdl_path()> or
    #   // <add_hdl_path_slice()> method.
    #   //
    #   extern function void configure (uvm_reg_block parent,
    #                                   string        hdl_path = "")
    def configure(self, parent, hdl_path=""):

        if parent is None:
            uvm_fatal("REG/None_PARENT","configure: parent argument is None")
        self.m_parent = parent

        if (self.m_access != "RW" and self.m_access != "RO"):
            uvm_error("RegModel", "Memory '" + self.get_full_name() + "' can only be RW or RO")
            self.m_access = "RW"

        cfg = UVMMemMamCfg()  # uvm_mem_mam_cfg
        cfg.n_bytes      = int((self.m_n_bits-1) / 8) + 1
        cfg.start_offset = 0
        cfg.end_offset   = self.m_size-1
        cfg.mode     = UVMMemMam.GREEDY
        cfg.locality = UVMMemMam.BROAD

        self.mam = UVMMemMam(self.get_full_name(), cfg, self)
        self.m_parent.add_mem(self)

        if hdl_path != "":
            self.add_hdl_path_slice(hdl_path, -1, -1)
        #endfunction: configure


    #
    #
    #   // Function: set_offset
    #   //
    #   // Modify the offset of the memory
    #   //
    #   // The offset of a memory within an address map is set using the
    #   // <uvm_reg_map::add_mem()> method.
    #   // This method is used to modify that offset dynamically.
    #   //
    #   // Note: Modifying the offset of a memory will make the abstract model
    #   // diverge from the specification that was used to create it.
    #   //
    #   extern virtual function void set_offset (uvm_reg_map    map,
    #                                            uvm_reg_addr_t offset,
    #                                            bit            unmapped = 0)
    #

    #
    #   /*local*/ extern virtual function void set_parent(uvm_reg_block parent)

    #   /*local*/ extern function void add_map(uvm_reg_map map)
    def add_map(self, _map):
        self.m_maps[_map] = 1

    #   /*local*/ extern function void Xlock_modelX()
    def Xlock_modelX(self):
        self.m_locked = 1

    #   /*local*/ extern function void Xadd_vregX(uvm_vreg vreg)
    def Xadd_vregX(self, vreg):
        self.m_vregs[vreg] = 1


    #   /*local*/ extern function void Xdelete_vregX(uvm_vreg vreg)


    #   //---------------------
    #   // Group: Introspection
    #   //---------------------
    #
    #   // Function: get_name
    #   //
    #   // Get the simple name
    #   //
    #   // Return the simple object name of this memory.
    #   //


    #   // Function: get_full_name
    #   //
    #   // Get the hierarchical name
    #   //
    #   // Return the hierarchal name of this memory.
    #   // The base of the hierarchical name is the root block.
    #   //
    #   extern virtual function string get_full_name()
    def get_full_name(self):
        if self.m_parent is None:
            return self.get_name()

        return self.m_parent.get_full_name() + "." + self.get_name()


    #   // Function: get_parent
    #   //
    #   // Get the parent block
    #   //
    #   extern virtual function uvm_reg_block get_parent ()
    def get_parent(self):
        return self.get_block()

    #   extern virtual function uvm_reg_block get_block  ()
    def get_block(self):
        return self.m_parent

    #
    #   // Function: get_n_maps
    #   //
    #   // Returns the number of address maps this memory is mapped in
    #   //
    #   extern virtual function int get_n_maps ()
    def get_n_maps(self):
        return self.m_maps.num()



    #   // Function: is_in_map
    #   //
    #   // Return TRUE if this memory is in the specified address ~map~
    #   //
    #   extern function bit is_in_map (uvm_reg_map map)
    #
    #
    #   // Function: get_maps
    #   //
    #   // Returns all of the address ~maps~ where this memory is mapped
    #   //
    #   extern virtual function void get_maps (ref uvm_reg_map maps[$])
    def get_maps(self, maps):
        for map_ in self.m_maps.key_list():
            maps.append(map_)


    #   /*local*/ extern function uvm_reg_map get_local_map   (uvm_reg_map map,
    #                                                          string caller = "")
    def get_local_map(self, _map, caller=""):
        if (_map is None):
            return self.get_default_map()
        if _map in self.m_maps:
            return _map

        for l in self.m_maps:
            local_map = l  # uvm_reg_map
            parent_map = local_map.get_parent_map()

            while parent_map is not None:
                if (parent_map == _map):
                    return local_map
                parent_map = parent_map.get_parent_map()

        cname = ""
        if caller != "":
            cname = " (called from " + caller + ")"
        uvm_warning("RegModel",
            "Memory '" + self.get_full_name() + "' is not contained within map '"
            + _map.get_full_name() + "'" + cname)
        return None
        #endfunction


    #   /*local*/ extern function uvm_reg_map get_default_map (string caller = "")
    def get_default_map(self, caller=""):

        # if mem is not associated with any may, return ~None~
        if self.m_maps.num() == 0:
            cname = ""
            if caller != "":
                cname = " (called from " + caller + ")"
            uvm_warning("RegModel",
                    "Memory '" + self.get_full_name() + "' is not registered with any map"
                    + cname)
            return None


        # if only one map, choose that
        default_map = None
        if self.m_maps.num() == 1:
            get_default_map = self.m_maps.first()

        # try to choose one based on default_map in parent blocks.
        for l in self.m_maps:
            _map = l  # uvm_reg_map
            blk = _map.get_parent()  # uvm_reg_block
            default_map = blk.get_default_map()  # uvm_reg_map
            if default_map is not None:
                local_map = self.get_local_map(default_map)
                if local_map is not None:
                    return local_map

        # if that fails, choose the first in self mem's maps
        get_default_map = self.m_maps.first()
        return get_default_map


    #   // Function: get_rights
    #   //
    #   // Returns the access rights of this memory.
    #   //
    #   // Returns "RW", "RO" or "WO".
    #   // The access rights of a memory is always "RW",
    #   // unless it is a shared memory
    #   // with access restriction in a particular address map.
    #   //
    #   // If no address map is specified and the memory is mapped in only one
    #   // address map, that address map is used. If the memory is mapped
    #   // in more than one address map, the default address map of the
    #   // parent block is used.
    #   //
    #   // If an address map is specified and
    #   // the memory is not mapped in the specified
    #   // address map, an error message is issued
    #   // and "RW" is returned.
    #   //
    #   extern virtual function string get_rights (uvm_reg_map map = None)
    def get_rights(self, _map=None):
        #
        info = None  # uvm_reg_map_info

        # No right restrictions if not shared
        if (self.m_maps.num() <= 1):
            return "RW"

        _map = self.get_local_map(_map,"get_rights()")
        if _map is None:
            return "RW"

        info = _map.get_mem_map_info(self)
        return info.rights
        #endfunction: get_rights



    #   // Function: get_access
    #   //
    #   // Returns the access policy of the memory when written and read
    #   // via an address map.
    #   //
    #   // If the memory is mapped in more than one address map,
    #   // an address ~map~ must be specified.
    #   // If access restrictions are present when accessing a memory
    #   // through the specified address map, the access mode returned
    #   // takes the access restrictions into account.
    #   // For example, a read-write memory accessed
    #   // through a domain with read-only restrictions would return "RO".
    #   //
    #   extern virtual function string get_access(uvm_reg_map map = None)
    def get_access(self, _map=None):
        get_access = self.m_access
        if (self.get_n_maps() == 1):
            return get_access
        _map = self.get_local_map(_map, "get_access()")
        if _map is None:
            return get_access

        # // Is the memory restricted in this map?
        rights = self.get_rights(_map)
        if rights == "RW":
            # // No restrictions
            return get_access
        elif rights == "RO":
            if get_access in ["RW", "RO"]:
                get_access = "RO"
            elif get_access == "WO":
                uvm_error("RegModel", "WO memory '" + self.get_full_name()
                    + "' restricted to RO in map '" + _map.get_full_name() + "'")
            else:
                uvm_error("RegModel", "Memory '" + self.get_full_name()
                    + "' has invalid access mode, '" + get_access + "'")

        elif rights == "WO":
            if get_access in ["RW", "WO"]:
                get_access = "WO"
            elif get_access == "RO":
                uvm_error("RegModel", "RO memory '" + self.get_full_name()
                    + "' restricted to WO in map '" + _map.get_full_name() + "'")
            else:
                uvm_error("RegModel", "Memory '" + self.get_full_name()
                    + "' has invalid access mode, '" + get_access + "'")
        else:
            uvm_error("RegModel", "Shared memory '" + self.get_full_name()
                + "' is not shared in map '" + _map.get_full_name() + "'")

    #   // Function: get_size
    #   //
    #   // Returns the number of unique memory locations in this memory.
    #   //
    #   extern function longint unsigned get_size()
    def get_size(self):
        return self.m_size

    #   // Function: get_n_bytes
    #   //
    #   // Return the width, in number of bytes, of each memory location
    #   //
    #   extern function int unsigned get_n_bytes()
    def get_n_bytes(self):
        return int((self.m_n_bits - 1) / 8) + 1


    #   // Function: get_n_bits
    #   //
    #   // Returns the width, in number of bits, of each memory location
    #   //
    #   extern function int unsigned get_n_bits()
    def get_n_bits(self):
        return self.m_n_bits

    #   // Function: get_max_size
    #   //
    #   // Returns the maximum width, in number of bits, of all memories
    #   //
    @classmethod
    def get_max_size(cls):
        return UVMMem.m_max_size


    #   // Function: get_virtual_registers
    #   //
    #   // Return the virtual registers in this memory
    #   //
    #   // Fills the specified array with the abstraction class
    #   // for all of the virtual registers implemented in this memory.
    #   // The order in which the virtual registers are located in the array
    #   // is not specified.
    #   //
    #   extern virtual function void get_virtual_registers(ref uvm_vreg regs[$])
    #
    #
    #   // Function: get_virtual_fields
    #   //
    #   // Return  the virtual fields in the memory
    #   //
    #   // Fills the specified dynamic array with the abstraction class
    #   // for all of the virtual fields implemented in this memory.
    #   // The order in which the virtual fields are located in the array is
    #   // not specified.
    #   //
    #   extern virtual function void get_virtual_fields(ref uvm_vreg_field fields[$])
    #
    #
    #   // Function: get_vreg_by_name
    #   //
    #   // Find the named virtual register
    #   //
    #   // Finds a virtual register with the specified name
    #   // implemented in this memory and returns
    #   // its abstraction class instance.
    #   // If no virtual register with the specified name is found, returns ~None~.
    #   //
    #   extern virtual function uvm_vreg get_vreg_by_name(string name)
    #
    #
    #   // Function: get_vfield_by_name
    #   //
    #   // Find the named virtual field
    #   //
    #   // Finds a virtual field with the specified name
    #   // implemented in this memory and returns
    #   // its abstraction class instance.
    #   // If no virtual field with the specified name is found, returns ~None~.
    #   //
    #   extern virtual function uvm_vreg_field  get_vfield_by_name(string name)
    #
    #
    #   // Function: get_vreg_by_offset
    #   //
    #   // Find the virtual register implemented at the specified offset
    #   //
    #   // Finds the virtual register implemented in this memory
    #   // at the specified ~offset~ in the specified address ~map~
    #   // and returns its abstraction class instance.
    #   // If no virtual register at the offset is found, returns ~None~.
    #   //
    #   extern virtual function uvm_vreg get_vreg_by_offset(uvm_reg_addr_t offset,
    #                                                       uvm_reg_map    map = None)
    #
    #
    #   // Function: get_offset
    #   //
    #   // Returns the base offset of a memory location
    #   //
    #   // Returns the base offset of the specified location in this memory
    #   // in an address ~map~.
    #   //
    #   // If no address map is specified and the memory is mapped in only one
    #   // address map, that address map is used. If the memory is mapped
    #   // in more than one address map, the default address map of the
    #   // parent block is used.
    #   //
    #   // If an address map is specified and
    #   // the memory is not mapped in the specified
    #   // address map, an error message is issued.
    #   //
    #   extern virtual function uvm_reg_addr_t  get_offset (uvm_reg_addr_t offset = 0,
    #                                                       uvm_reg_map    map = None)
    #
    #
    #   // Function: get_address
    #   //
    #   // Returns the base external physical address of a memory location
    #   //
    #   // Returns the base external physical address of the specified location
    #   // in this memory if accessed through the specified address ~map~.
    #   //
    #   // If no address map is specified and the memory is mapped in only one
    #   // address map, that address map is used. If the memory is mapped
    #   // in more than one address map, the default address map of the
    #   // parent block is used.
    #   //
    #   // If an address map is specified and
    #   // the memory is not mapped in the specified
    #   // address map, an error message is issued.
    #   //
    #   extern virtual function uvm_reg_addr_t  get_address(uvm_reg_addr_t  offset = 0,
    #                                                       uvm_reg_map   map = None)
    #
    #
    #   // Function: get_addresses
    #   //
    #   // Identifies the external physical address(es) of a memory location
    #   //
    #   // Computes all of the external physical addresses that must be accessed
    #   // to completely read or write the specified location in this memory.
    #   // The addressed are specified in little endian order.
    #   // Returns the number of bytes transferred on each access.
    #   //
    #   // If no address map is specified and the memory is mapped in only one
    #   // address map, that address map is used. If the memory is mapped
    #   // in more than one address map, the default address map of the
    #   // parent block is used.
    #   //
    #   // If an address map is specified and
    #   // the memory is not mapped in the specified
    #   // address map, an error message is issued.
    #   //
    #   extern virtual function int get_addresses(uvm_reg_addr_t     offset = 0,
    #                                             uvm_reg_map        map=None,
    #                                             ref uvm_reg_addr_t addr[])
    #

    #   //------------------
    #   // Group: HDL Access
    #   //------------------


    #   // Task: write
    #   //
    #   // Write the specified value in a memory location
    #   //
    #   // Write ~value~ in the memory location that corresponds to this
    #   // abstraction class instance at the specified ~offset~
    #   // using the specified access ~path~.
    #   // If the memory is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   // If a back-door access path is used, the effect of writing
    #   // the register through a physical access is mimicked. For
    #   // example, a read-only memory will not be written.
    #   //
    #   extern virtual task write(output uvm_status_e       status,
    #                             input  uvm_reg_addr_t     offset,
    #                             input  uvm_reg_data_t     value,
    #                             input  uvm_path_e         path   = UVM_DEFAULT_PATH,
    #                             input  uvm_reg_map        map = None,
    #                             input  uvm_sequence_base  parent = None,
    #                             input  int                prior = -1,
    #                             input  uvm_object         extension = None,
    #                             input  string             fname = "",
    #                             input  int                lineno = 0)
    async def write(self, status, offset, value, path=UVM_DEFAULT_PATH, _map=None, parent=None,
            prior=-1, extension=None, fname="", lineno=0):
        uvm_check_output_args([status])

        # create an abstract transaction for this operation
        rw = UVMRegItem.type_id.create("mem_write",None,self.get_full_name())
        rw.element      = self
        rw.element_kind = UVM_MEM
        rw.kind         = UVM_WRITE
        rw.offset       = offset
        rw.value[0]     = value
        rw.path         = path
        rw.map          = _map
        rw.parent       = parent
        rw.prior        = prior
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno

        await self.do_write(rw)
        status.append(rw.status)


    #   // Task: read
    #   //
    #   // Read the current value from a memory location
    #   //
    #   // Read and return ~value~ from the memory location that corresponds to this
    #   // abstraction class instance at the specified ~offset~
    #   // using the specified access ~path~.
    #   // If the register is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   //
    #   extern virtual task read(output uvm_status_e        status,
    #                            input  uvm_reg_addr_t      offset,
    #                            output uvm_reg_data_t      value,
    #                            input  uvm_path_e          path   = UVM_DEFAULT_PATH,
    #                            input  uvm_reg_map         map = None,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  int                 prior = -1,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)
    async def read(self, status, offset, value, path=UVM_DEFAULT_PATH, _map=None,
            parent=None, prior=-1, extension=None, fname="", lineno=0):
        uvm_check_output_args([status, value])

        rw = UVMRegItem.type_id.create("mem_read",None,self.get_full_name())
        rw.element      = self
        rw.element_kind = UVM_MEM
        rw.kind         = UVM_READ
        rw.value[0]     = 0
        rw.offset       = offset
        rw.path         = path
        rw.map          = _map
        rw.parent       = parent
        rw.prior        = prior
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno

        await self.do_read(rw)

        status.append(rw.status)
        value.append(rw.value[0])



    #   // Task: burst_write
    #   //
    #   // Write the specified values in memory locations
    #   //
    #   // Burst-write the specified ~values~ in the memory locations
    #   // beginning at the specified ~offset~.
    #   // If the memory is mapped in more than one address map,
    #   // an address ~map~ must be specified if not using the backdoor.
    #   // If a back-door access path is used, the effect of writing
    #   // the register through a physical access is mimicked. For
    #   // example, a read-only memory will not be written.
    #   //
    #   extern virtual task burst_write(output uvm_status_e      status,
    #                                   input  uvm_reg_addr_t    offset,
    #                                   input  uvm_reg_data_t    value[],
    #                                   input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                                   input  uvm_reg_map       map = None,
    #                                   input  uvm_sequence_base parent = None,
    #                                   input  int               prior = -1,
    #                                   input  uvm_object        extension = None,
    #                                   input  string            fname = "",
    #                                   input  int               lineno = 0)
    async def burst_write(self, status, offset, value, path=UVM_DEFAULT_PATH,
        _map=None, parent=None, prior=-1, extension=None, fname="",
            lineno=0):
        uvm_check_output_args([status])
        rw = UVMRegItem.type_id.create("mem_burst_write", None, self.get_full_name())
        rw.element      = self
        rw.element_kind = UVM_MEM
        rw.kind         = UVM_BURST_WRITE
        rw.offset       = offset
        rw.value        = value
        rw.path         = path
        rw.map          = _map
        rw.parent       = parent
        rw.prior        = prior
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno
        await self.do_write(rw)
        status.append(rw.status)



    #   // Task: burst_read
    #   //
    #   // Read values from memory locations
    #   //
    #   // Burst-read into ~values~ the data the memory locations
    #   // beginning at the specified ~offset~.
    #   // If the memory is mapped in more than one address map,
    #   // an address ~map~ must be specified if not using the backdoor.
    #   // If a back-door access path is used, the effect of writing
    #   // the register through a physical access is mimicked. For
    #   // example, a read-only memory will not be written.
    #   //
    #   extern virtual task burst_read(output uvm_status_e      status,
    #                                  input  uvm_reg_addr_t    offset,
    #                                  ref    uvm_reg_data_t    value[],
    #                                  input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                                  input  uvm_reg_map       map = None,
    #                                  input  uvm_sequence_base parent = None,
    #                                  input  int               prior = -1,
    #                                  input  uvm_object        extension = None,
    #                                  input  string            fname = "",
    #                                  input  int               lineno = 0)
    async def burst_read(self, status, offset,
    value, path=UVM_DEFAULT_PATH, _map=None, parent=None, prior=-1, extension=None, fname="",
            lineno=0):
        uvm_check_output_args([status, value])
        rw = UVMRegItem.type_id.create("mem_burst_read", None, self.get_full_name())
        rw.element      = self
        rw.element_kind = UVM_MEM
        rw.kind         = UVM_BURST_READ
        rw.offset       = offset
        rw.value        = value
        rw.path         = path
        rw.map          = _map
        rw.parent       = parent
        rw.prior        = prior
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno

        await self.do_read(rw)
        status.append(rw.status)
        for val in rw.value:
            value.append(val)


    #   // Task: poke
    #   //
    #   // Deposit the specified value in a memory location
    #   //
    #   // Deposit the value in the DUT memory location corresponding to this
    #   // abstraction class instance at the specified ~offset~, as-is,
    #   // using a back-door access.
    #   //
    #   // Uses the HDL path for the design abstraction specified by ~kind~.
    #   //
    #   extern virtual task poke(output uvm_status_e       status,
    #                            input  uvm_reg_addr_t     offset,
    #                            input  uvm_reg_data_t     value,
    #                            input  string             kind = "",
    #                            input  uvm_sequence_base  parent = None,
    #                            input  uvm_object         extension = None,
    #                            input  string             fname = "",
    #                            input  int                lineno = 0)
    def poke(self, status,
            offset, value, kind="", parent=None, extension=None, fname="", lineno=0):
        uvm_check_output_args([status])
        bkdr = self.get_backdoor()  # uvm_reg_backdoor
        rw = None  # uvm_reg_item

        self.m_fname = fname
        self.m_lineno = lineno

        if bkdr is None and self.has_hdl_path(kind) is False:
            uvm_error("RegModel", "No backdoor access available in memory '"
                      + self.get_full_name() + "'")
            status.append(UVM_NOT_OK)
            return

        # create an abstract transaction for this operation
        rw = UVMRegItem.type_id.create("mem_poke_item", None, self.get_full_name())
        rw.element      = self
        rw.path         = UVM_BACKDOOR
        rw.element_kind = UVM_MEM
        rw.kind         = UVM_READ
        rw.offset       = offset
        rw.value[0]     = value & ((1 << self.m_n_bits)-1)
        rw.bd_kind      = kind
        rw.parent       = parent
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno

        if bkdr is not None:
            bkdr.write(rw)
        else:
            self.backdoor_write(rw)

        status.append(rw.status)

        uvm_info("RegModel", sv.sformatf("Poked memory '%s[%0d]' with value 'h%h",
            self.get_full_name(), offset, value), UVM_HIGH)


    #   // Task: peek
    #   //
    #   // Read the current value from a memory location
    #   //
    #   // Sample the value in the DUT memory location corresponding to this
    #   // abstraction class instance at the specified ~offset~
    #   // using a back-door access.
    #   // The memory location value is sampled, not modified.
    #   //
    #   // Uses the HDL path for the design abstraction specified by ~kind~.
    #   //
    #   extern virtual task peek(output uvm_status_e       status,
    #                            input  uvm_reg_addr_t     offset,
    #                            output uvm_reg_data_t     value,
    #                            input  string             kind = "",
    #                            input  uvm_sequence_base  parent = None,
    #                            input  uvm_object         extension = None,
    #                            input  string             fname = "",
    #                            input  int                lineno = 0)
    def peek(self, status, offset, value, kind="", parent=None, extension=None, fname="",
            lineno=0):
        uvm_check_output_args([status, value])
        bkdr = self.get_backdoor()  # uvm_reg_backdoor
        rw = None  # uvm_reg_item

        self.m_fname = fname
        self.m_lineno = lineno

        if bkdr is None and self.has_hdl_path(kind) is False:
            uvm_error("RegModel", "No backdoor access available in memory '"
                      + self.get_full_name() + "'")
            status.append(UVM_NOT_OK)
            return

        # create an abstract transaction for this operation
        rw = UVMRegItem.type_id.create("mem_peek_item", None, self.get_full_name())
        rw.element      = self
        rw.path         = UVM_BACKDOOR
        rw.element_kind = UVM_MEM
        rw.kind         = UVM_READ
        rw.offset       = offset
        rw.bd_kind      = kind
        rw.parent       = parent
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno

        if bkdr is not None:
            bkdr.read(rw)
        else:
            self.backdoor_read(rw)

        status.append(rw.status)
        value.append(rw.value[0])

        uvm_info("RegModel", sv.sformatf("Peeked memory '%s[%0d]' has value '%s'",
            self.get_full_name(), offset, str(value)), UVM_HIGH)


    #   extern protected function bit Xcheck_accessX (input uvm_reg_item rw,
    #                                                 output uvm_reg_map_info map_info,
    #                                                 input string caller)
    def Xcheck_accessX(self, rw, map_info, caller):
        uvm_check_output_args([map_info])

        if rw.offset >= self.m_size:
            uvm_error(self.get_type_name(),
               sv.sformatf("Offset 'h%0h exceeds size of memory, 'h%0h",
                 rw.offset, self.m_size))
            rw.status = UVM_NOT_OK
            return 0

        if rw.path == UVM_DEFAULT_PATH:
            rw.path = self.m_parent.get_default_path()

        if rw.path == UVM_BACKDOOR:
           if self.get_backdoor() is None and  self.has_hdl_path() is False:
               uvm_warning("RegModel",
                 "No backdoor access available for memory '" + self.get_full_name()
                 + "' . Using frontdoor instead.")
               rw.path = UVM_FRONTDOOR
           else:
               rw.map = UVMRegMap.backdoor()

        if rw.path != UVM_BACKDOOR:
            rw.local_map = self.get_local_map(rw.map,caller)

            if rw.local_map is None:
                uvm_error(self.get_type_name(),
                    "No transactor available to physically access memory from map '"
                    + rw.map.get_full_name() + "'")
                rw.status = UVM_NOT_OK
                return 0

            _map_info_data = rw.local_map.get_mem_map_info(self)
            map_info.append(_map_info_data)

            if _map_info_data.frontdoor is None:

                if (_map_info_data.unmapped):
                    uvm_error("RegModel", "Memory '" + self.get_full_name()
                              + "' unmapped in map '" + rw.map.get_full_name()
                              + "' and does not have a user-defined frontdoor")
                    rw.status = UVM_NOT_OK
                    return 0

                if len(rw.value) > 1:
                    if self.get_n_bits() > rw.local_map.get_n_bytes()*8:
                        uvm_error("RegModel",
                             sv.sformatf("Cannot burst a %0d-bit memory through a narrower data path (%0d bytes)",
                             self.get_n_bits(), rw.local_map.get_n_bytes()*8))
                        rw.status = UVM_NOT_OK
                        return 0

                    if rw.offset + len(rw.value) > self.m_size:
                        uvm_error("RegModel",
                           sv.sformatf("Burst of size 'd%0d starting at offset 'd%0d exceeds size of memory, 'd%0d",
                               len(rw.value), rw.offset, self.m_size))
                        return 0

            if rw.map is None:
                rw.map = rw.local_map

        return 1


    #   extern virtual task do_write (uvm_reg_item rw)
    async def do_write(self, rw):  # task
        cbs = UVMMemCbIter(self)
        arr_map_info = []  # uvm_reg_map_info

        self.m_fname  = rw.fname
        self.m_lineno = rw.lineno

        if self.Xcheck_accessX(rw, arr_map_info, "burst_write()") is False:
            await uvm_zero_delay()
            return

        self.m_write_in_progress = True
        rw.status = UVM_IS_OK

        # PRE-WRITE CBS
        await self.pre_write(rw)
        #   for (uvm_reg_cbs cb=cbs.first(); cb is not None; cb=cbs.next())
        #      cb.pre_write(rw)
        cb = cbs.first()
        while cb is not None:
            await cb.pre_write(rw)
            cb = cbs.next()

        if rw.status != UVM_IS_OK:
            self.m_write_in_progress = False
            return

        rw.status = UVM_NOT_OK

        # FRONTDOOR
        if rw.path == UVM_FRONTDOOR:
            map_info = arr_map_info[0]

            system_map = rw.local_map.get_root_map()
            if map_info.frontdoor is not None:
                fd = map_info.frontdoor  # uvm_reg_frontdoor
                fd.rw_info = rw
                if fd.sequencer is None:
                    fd.sequencer = system_map.get_sequencer()
                await fd.start(fd.sequencer, rw.parent)
            else:
                await rw.local_map.do_write(rw)

            if rw.status != UVM_NOT_OK:
                _min = rw.offset
                _max = rw.offset + len(rw.value) + 1
                for idx in range(_min, _max):
                    self.XsampleX(map_info.mem_range.stride * idx, 0, rw.map)
                    self.m_parent.XsampleX(map_info.offset +
                            (map_info.mem_range.stride * idx), 0, rw.map)

        else:
            #raise Exception("Backdoor access not implemented yet for UVMMem!")
            # // BACKDOOR TODO
            # // Mimick front door access, i.e. do not write read-only memories
            if self.get_access(rw.map) == "RW":
                bkdr = self.get_backdoor()  # uvm_reg_backdoor
                if bkdr is not None:
                    await bkdr.write(rw)
                else:
                    # await self.backdoor_write(rw)
                    self.backdoor_write(rw)
            else:
                rw.status = UVM_IS_OK

        #   // POST-WRITE CBS
        await self.post_write(rw)
        cb = cbs.first()
        while cb is not None:
            await cb.post_write(rw)
            cb = cbs.next()

        # REPORT
        if uvm_report_enabled(UVM_HIGH, UVM_INFO, "RegModel"):
            path_s = ""
            value_s = ""
            pre_s = ""
            range_s = ""
            if rw.path == UVM_FRONTDOOR:
                path_s = "map " + rw.map.get_full_name()
                if map_info.frontdoor is not None:
                    path_s = "user frontdoor"
            else:
                path_s = "DPI backdoor"
                if self.get_backdoor() is not None:
                    path_s = "user backdoor"

            if len(rw.value) > 1:
                value_s = "='{"
                pre_s = "Burst "
                for i in range(len(rw.value)):
                    value_s = value_s + sv.sformatf("%0h,",rw.value[i])
                value_s = value_s[:-1] + "}"
                range_s = sv.sformatf("[%0d:%0d]",rw.offset,rw.offset+len(rw.value))
            else:
                value_s = sv.sformatf("=%0h",rw.value[0])
                range_s = sv.sformatf("[%0d]",rw.offset)
            uvm_report_info("RegModel", pre_s + "Wrote memory via " + path_s + ": " +
                    self.get_full_name() + range_s + value_s, UVM_HIGH)
        self.m_write_in_progress = False


    #   extern virtual task do_read  (uvm_reg_item rw)
    async def do_read(self, rw):  # task
        cbs = UVMMemCbIter(self)
        arr_map_info = []  # uvm_reg_map_info

        self.m_fname = rw.fname
        self.m_lineno = rw.lineno

        if self.Xcheck_accessX(rw, arr_map_info, "burst_read()") is False:
            await uvm_zero_delay()
            return

        self.m_read_in_progress = True
        rw.status = UVM_IS_OK
        #
        # PRE-READ CBS
        await self.pre_read(rw)
        cb = cbs.first()
        while cb is not None:
            cb.pre_read(rw)
            cb = cbs.next()

        if rw.status != UVM_IS_OK:
            self.m_read_in_progress = False
            return

        rw.status = UVM_NOT_OK

        # FRONTDOOR
        if rw.path == UVM_FRONTDOOR:
            map_info = arr_map_info[0]
            system_map = rw.local_map.get_root_map()

            if (map_info.frontdoor is not None):
                fd = map_info.frontdoor  # uvm_reg_frontdoor
                fd.rw_info = rw
                if fd.sequencer is None:
                    fd.sequencer = system_map.get_sequencer()
                await fd.start(fd.sequencer, rw.parent)
            else:
                await rw.local_map.do_read(rw)

            if rw.status != UVM_NOT_OK:
                _min = rw.offset
                _max = rw.offset + len(rw.value) + 1
                for idx in range(_min, _max):
                    self.XsampleX(map_info.mem_range.stride * idx, 1, rw.map)
                    self.m_parent.XsampleX(map_info.offset +
                            (map_info.mem_range.stride * idx), 1, rw.map)
        # // BACKDOOR
        else:
            bkdr = self.get_backdoor()  # uvm_reg_backdoor
            if bkdr is not None:
                await bkdr.read(rw)
            else:
                # await self.backdoor_read(rw)
                self.backdoor_read(rw)

        # POST-READ CBS
        await self.post_read(rw)
        cb = cbs.first()
        while cb is not None:
            await cb.post_read(rw)
            cb = cbs.next()

        # REPORT
        if uvm_report_enabled(UVM_HIGH, UVM_INFO, "RegModel"):
            path_s = ""
            value_s = ""
            pre_s = ""
            range_s = ""
            if rw.path == UVM_FRONTDOOR:
                path_s = "map " + rw.map.get_full_name()
                map_info = arr_map_info[0]
                if map_info.frontdoor is not None:
                    path_s = "user frontdoor"
            else:
                path_s = "DPI backdoor"
                if self.get_backdoor() is not None:
                    path_s = "user backdoor"

            if len(rw.value) > 1:
                value_s = "='{"
                pre_s = "Burst "
                for i in range(len(rw.value)):
                    value_s = value_s + sv.sformatf("%0h,", int(rw.value[i]))
                value_s = value_s[:-1] + "}"
                range_s = sv.sformatf("[%0d:%0d]",rw.offset,rw.offset+len(rw.value))
            else:
                value_s = sv.sformatf("=%0h", int(rw.value[0]))
                range_s = sv.sformatf("[%0d]", rw.offset)

            uvm_report_info("RegModel", pre_s + "Read memory via " + path_s + ": "
                    + self.get_full_name() + range_s + value_s, UVM_HIGH)

        self.m_read_in_progress = False
        #
        #endtask: do_read


    #   //-----------------
    #   // Group: Frontdoor
    #   //-----------------


    #   // Function: set_frontdoor
    #   //
    #   // Set a user-defined frontdoor for this memory
    #   //
    #   // By default, memories are mapped linearly into the address space
    #   // of the address maps that instantiate them.
    #   // If memories are accessed using a different mechanism,
    #   // a user-defined access
    #   // mechanism must be defined and associated with
    #   // the corresponding memory abstraction class
    #   //
    #   // If the memory is mapped in multiple address maps, an address ~map~
    #   // must be specified.
    #   //
    #   extern function void set_frontdoor(uvm_reg_frontdoor ftdr,
    #                                      uvm_reg_map map = None,
    #                                      string fname = "",
    #                                      int lineno = 0)

    #
    #
    #   // Function: get_frontdoor
    #   //
    #   // Returns the user-defined frontdoor for this memory
    #   //
    #   // If ~None~, no user-defined frontdoor has been defined.
    #   // A user-defined frontdoor is defined
    #   // by using the <uvm_mem::set_frontdoor()> method.
    #   //
    #   // If the memory is mapped in multiple address maps, an address ~map~
    #   // must be specified.
    #   //
    #   extern function uvm_reg_frontdoor get_frontdoor(uvm_reg_map map = None)

    #
    #
    #   //----------------
    #   // Group: Backdoor
    #   //----------------

    #
    #   // Function: set_backdoor
    #   //
    #   // Set a user-defined backdoor for this memory
    #   //
    #   // By default, memories are accessed via the built-in string-based
    #   // DPI routines if an HDL path has been specified using the
    #   // <uvm_mem::configure()> or <uvm_mem::add_hdl_path()> method.
    #   // If this default mechanism is not suitable (e.g. because
    #   // the memory is not implemented in pure SystemVerilog)
    #   // a user-defined access
    #   // mechanism must be defined and associated with
    #   // the corresponding memory abstraction class
    #   //
    #   extern function void set_backdoor (uvm_reg_backdoor bkdr,
    #                                      string fname = "",
    #                                      int lineno = 0)
    def set_backdoor(self, bkdr, fname="", lineno=0):
        self.m_fname = fname
        self.m_lineno = lineno
        self.m_backdoor = bkdr


    #   // Function: get_backdoor
    #   //
    #   // Returns the user-defined backdoor for this memory
    #   //
    #   // If ~None~, no user-defined backdoor has been defined.
    #   // A user-defined backdoor is defined
    #   // by using the <uvm_reg::set_backdoor()> method.
    #   //
    #   // If ~inherit~ is TRUE, returns the backdoor of the parent block
    #   // if none have been specified for this memory.
    #   //
    #   extern function uvm_reg_backdoor get_backdoor(bit inherited = 1)
    def get_backdoor(self, inherited=1):
        #
        if self.m_backdoor is None and inherited:
            blk = self.get_parent()  # uvm_reg_block
            bkdr = None  # uvm_reg_backdoor
            while blk is not None:
                bkdr = blk.get_backdoor()
                if bkdr is not None:
                    self.m_backdoor = bkdr
                    break
                blk = blk.get_parent()
        #
        return self.m_backdoor
        #endfunction: get_backdoor



    #   // Function: clear_hdl_path
    #   //
    #   // Delete HDL paths
    #   //
    #   // Remove any previously specified HDL path to the memory instance
    #   // for the specified design abstraction.
    #   //
    #   extern function void clear_hdl_path (string kind = "RTL")

    #
    #
    #   // Function: add_hdl_path
    #   //
    #   // Add an HDL path
    #   //
    #   // Add the specified HDL path to the memory instance for the specified
    #   // design abstraction. This method may be called more than once for the
    #   // same design abstraction if the memory is physically duplicated
    #   // in the design abstraction
    #   //
    #   extern function void add_hdl_path (uvm_hdl_path_slice slices[],
    #                                      string kind = "RTL")
    #
    #

    #   // Function: add_hdl_path_slice
    #   //
    #   // Add the specified HDL slice to the HDL path for the specified
    #   // design abstraction.
    #   // If ~first~ is TRUE, starts the specification of a duplicate
    #   // HDL implementation of the memory.
    #   //
    #   extern function void add_hdl_path_slice(string name,
    #                                           int offset,
    #                                           int size,
    #                                           bit first = 0,
    #                                           string kind = "RTL")
    def add_hdl_path_slice(self, name, offset, size, first=0, kind="RTL"):
        paths = self.m_hdl_paths_pool.get(kind)  # #    uvm_queue #(uvm_hdl_path_concat)
        concat = None  #    uvm_hdl_path_concat

        if first or paths.size() == 0:
            concat = uvm_hdl_path_concat()
            paths.push_back(concat)
        else:
            concat = paths.get(paths.size()-1)

        concat.add_path(name, offset, size)
    #endfunction


    #
    #
    #   // Function: has_hdl_path
    #   //
    #   // Check if a HDL path is specified
    #   //
    #   // Returns TRUE if the memory instance has a HDL path defined for the
    #   // specified design abstraction. If no design abstraction is specified,
    #   // uses the default design abstraction specified for the parent block.
    #   //
    #   extern function bit  has_hdl_path (string kind = "")
    def has_hdl_path(self, kind=""):
        if kind == "":
            kind = self.m_parent.get_default_hdl_path()
        return self.m_hdl_paths_pool.exists(kind)
        #endfunction


    #   // Function: get_hdl_path
    #   //
    #   // Get the incremental HDL path(s)
    #   //
    #   // Returns the HDL path(s) defined for the specified design abstraction
    #   // in the memory instance.
    #   // Returns only the component of the HDL paths that corresponds to
    #   // the memory, not a full hierarchical path
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for the parent block is used.
    #   //
    #   extern function void get_hdl_path (ref uvm_hdl_path_concat paths[$],
    #                                      input string kind = "")

    #
    #
    #   // Function: get_full_hdl_path
    #   //
    #   // Get the full hierarchical HDL path(s)
    #   //
    #   // Returns the full hierarchical HDL path(s) defined for the specified
    #   // design abstraction in the memory instance.
    #   // There may be more than one path returned even
    #   // if only one path was defined for the memory instance, if any of the
    #   // parent components have more than one path defined for the same design
    #   // abstraction
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for each ancestor block is used to get each incremental path.
    #   //
    #   extern function void get_full_hdl_path (ref uvm_hdl_path_concat paths[$],
    #                                           input string kind = "",
    #                                           input string separator = ".")
    def get_full_hdl_path(self, paths, kind="", separator="."):

        if kind == "":
            kind = self.m_parent.get_default_hdl_path()

        if self.has_hdl_path(kind) is False:
            uvm_error("RegModel",
                "Memory does not have hdl path defined for abstraction '" + kind + "'")
            return

        hdl_paths = self.m_hdl_paths_pool.get(kind)
        parent_paths = []  # string[$]
        self.m_parent.get_full_hdl_path(parent_paths, kind, separator)

        for i in range(len(hdl_paths)):
            hdl_concat = hdl_paths.get(i)

            for j in range(len(parent_paths)):
                t = uvm_hdl_path_concat()

                for k in range(len(hdl_concat.slices)):
                    if hdl_concat.slices[k].path == "":
                        t.add_path(parent_paths[j])
                    else:
                        t.add_path(parent_paths[j] + separator
                            + hdl_concat.slices[k].path,
                                  hdl_concat.slices[k].offset,
                                  hdl_concat.slices[k].size)
                paths.append(t)


    #   // Function: get_hdl_path_kinds
    #   //
    #   // Get design abstractions for which HDL paths have been defined
    #   //
    #   extern function void get_hdl_path_kinds (ref string kinds[$])


    #   // Function: backdoor_read
    #   //
    #   // User-define backdoor read access
    #   //
    #   // Override the default string-based DPI backdoor access read
    #   // for this memory type.
    #   // By default calls <uvm_mem::backdoor_read_func()>.
    #   //
    #   extern virtual protected task backdoor_read(uvm_reg_item rw)
    def backdoor_read(self, rw):
        rw.status = self.backdoor_read_func(rw)


    #   // Function: backdoor_write
    #   //
    #   // User-defined backdoor read access
    #   //
    #   // Override the default string-based DPI backdoor access write
    #   // for this memory type.
    #   //
    #   extern virtual task backdoor_write(uvm_reg_item rw)
    def backdoor_write(self, rw):
        paths = []
        ok = 1
        self.get_full_hdl_path(paths,rw.bd_kind)

        for mem_idx in range(len(rw.value)):
            idx = str(rw.offset + mem_idx)
            for i in range(len(paths)):
                hdl_concat = paths[i]  # uvm_hdl_path_concat
                for j in range(len(hdl_concat.slices)):
                    uvm_info("RegModel", sv.sformatf("backdoor_write to %s ", hdl_concat.slices[j].path),UVM_DEBUG)

                    if (hdl_concat.slices[j].offset < 0):
                        ok &= uvm_hdl.uvm_hdl_deposit(hdl_concat.slices[j].path
                                + "[" + idx + "]", rw.value[mem_idx])
                        continue

                    _slice = 0
                    _slice = rw.value[mem_idx] >> hdl_concat.slices[j].offset
                    _slice &= (1 << hdl_concat.slices[j].size)-1
                    ok &= uvm_hdl.uvm_hdl_deposit(hdl_concat.slices[j].path +
                            "[" + idx + "]", _slice)
        rw.status = UVM_NOT_OK
        if ok:
            rw.status = UVM_IS_OK
        #endtask


    #   // Function: backdoor_read_func
    #   //
    #   // User-defined backdoor read access
    #   //
    #   // Override the default string-based DPI backdoor access read
    #   // for this memory type.
    #   //
    #   extern virtual function uvm_status_e backdoor_read_func(uvm_reg_item rw)
    def backdoor_read_func(self, rw):

        paths = []  # uvm_hdl_path_concat [$]
        val = 0x0
        ok = 1

        self.get_full_hdl_path(paths, rw.bd_kind)

        # foreach (rw.value[mem_idx]):
        for mem_idx in range(len(rw.value)):
            idx = str(rw.offset + mem_idx)
            for i in range(len(paths)):
                hdl_concat = paths[i]  # uvm_hdl_path_concat
                val = 0
                for j in range(len(hdl_concat.slices)):
                    hdl_path = hdl_concat.slices[j].path + "[" + idx + "]"
                    uvm_info("RegModel", "backdoor_read from " + hdl_path, UVM_DEBUG)

                    if hdl_concat.slices[j].offset < 0:
                        arr = []
                        ok &= uvm_hdl.uvm_hdl_read(hdl_path, arr)
                        val = arr[0]
                        continue

                    _slice = 0
                    k = hdl_concat.slices[j].offset
                    ok &= uvm_hdl.uvm_hdl_read(hdl_path, _slice)

                    # TODO Loops all bits one by one
                    for n in range(hdl_concat.slices[j].size):
                        slice_n_bit = (_slice >> n) & 0x1
                        # Need to set slice_n_bit in pos k
                        val |= (slice_n_bit << k) & (1 << k)
                        k += 1
                        # _slice >>= 1

                val &= (1 << self.m_n_bits)-1

                if i == 0:
                    rw.value[mem_idx] = val

                if val != rw.value[mem_idx]:
                    uvm_error("RegModel", sv.sformatf(ERR_MEM_BD_READ,
                       self.get_full_name(), rw.value[mem_idx], uvm_hdl_concat2string(paths[0]),
                       val, uvm_hdl_concat2string(paths[i])))
                    return UVM_NOT_OK

        rw.status = UVM_NOT_OK
        if ok:
            rw.status = UVM_IS_OK
        return rw.status
        #endfunction



    #   //-----------------
    #   // Group: Callbacks
    #   //-----------------
    #   `uvm_register_cb(uvm_mem, uvm_reg_cbs)


    #   // Task: pre_write
    #   //
    #   // Called before memory write.
    #   //
    #   // If the ~offset~, ~value~, access ~path~,
    #   // or address ~map~ are modified, the updated offset, data value,
    #   // access path or address map will be used to perform the memory operation.
    #   // If the ~status~ is modified to anything other than <UVM_IS_OK>,
    #   // the operation is aborted.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of this method.
    #   //
    async def pre_write(self, rw):
        await uvm_zero_delay()

    #   // Task: post_write
    #   //
    #   // Called after memory write.
    #   //
    #   // If the ~status~ is modified, the updated status will be
    #   // returned by the memory operation.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of this method.
    #   //
    #   virtual task post_write(uvm_reg_item rw); endtask
    async def post_write(self, rw):
        await uvm_zero_delay()


    #   // Task: pre_read
    #   //
    #   // Called before memory read.
    #   //
    #   // If the ~offset~, access ~path~ or address ~map~ are modified,
    #   // the updated offset, access path or address map will be used to perform
    #   // the memory operation.
    #   // If the ~status~ is modified to anything other than <UVM_IS_OK>,
    #   // the operation is aborted.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of this method.
    #   //
    #   virtual task pre_read(uvm_reg_item rw); endtask
    async def pre_read(self, rw):
        await uvm_zero_delay()


    #   // Task: post_read
    #   //
    #   // Called after memory read.
    #   //
    #   // If the readback data or ~status~ is modified,
    #   // the updated readback //data or status will be
    #   // returned by the memory operation.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of this method.
    #   //
    #   virtual task post_read(uvm_reg_item rw); endtask
    async def post_read(self, rw):
        await uvm_zero_delay()


    #   //----------------
    #   // Group: Coverage
    #   //----------------

    #   // Function: build_coverage
    #   //
    #   // Check if all of the specified coverage model must be built.
    #   //
    #   // Check which of the specified coverage model must be built
    #   // in this instance of the memory abstraction class,
    #   // as specified by calls to <uvm_reg::include_coverage()>.
    #   //
    #   // Models are specified by adding the symbolic value of individual
    #   // coverage model as defined in <uvm_coverage_model_e>.
    #   // Returns the sum of all coverage models to be built in the
    #   // memory model.
    #   //
    #   extern protected function uvm_reg_cvr_t build_coverage(uvm_reg_cvr_t models)
    def build_coverage(self, models):
        build_coverage = UVM_NO_COVERAGE
        cov_arr = []
        uvm_reg_cvr_rsrc_db.read_by_name("uvm_reg::" + self.get_full_name(),
                "include_coverage", cov_arr, self)
        build_coverage = cov_arr[0]
        return build_coverage & models


    #   // Function: add_coverage
    #   //
    #   // Specify that additional coverage models are available.
    #   //
    #   // Add the specified coverage model to the coverage models
    #   // available in this class.
    #   // Models are specified by adding the symbolic value of individual
    #   // coverage model as defined in <uvm_coverage_model_e>.
    #   //
    #   // This method shall be called only in the constructor of
    #   // subsequently derived classes.
    #   //
    #   extern virtual protected function void add_coverage(uvm_reg_cvr_t models)
    def add_coverage(self, models):
        self.m_has_cover |= models


    #   // Function: has_coverage
    #   //
    #   // Check if memory has coverage model(s)
    #   //
    #   // Returns TRUE if the memory abstraction class contains a coverage model
    #   // for all of the models specified.
    #   // Models are specified by adding the symbolic value of individual
    #   // coverage model as defined in <uvm_coverage_model_e>.
    #   //
    #   extern virtual function bit has_coverage(uvm_reg_cvr_t models)
    def has_coverage(self, models):
        return ((self.m_has_cover & models) == models)


    #   // Function: set_coverage
    #   //
    #   // Turns on coverage measurement.
    #   //
    #   // Turns the collection of functional coverage measurements on or off
    #   // for this memory.
    #   // The functional coverage measurement is turned on for every
    #   // coverage model specified using <uvm_coverage_model_e> symbolic
    #   // identifiers.
    #   // Multiple functional coverage models can be specified by adding
    #   // the functional coverage model identifiers.
    #   // All other functional coverage models are turned off.
    #   // Returns the sum of all functional
    #   // coverage models whose measurements were previously on.
    #   //
    #   // This method can only control the measurement of functional
    #   // coverage models that are present in the memory abstraction classes,
    #   // then enabled during construction.
    #   // See the <uvm_mem::has_coverage()> method to identify
    #   // the available functional coverage models.
    #   //
    #   extern virtual function uvm_reg_cvr_t set_coverage(uvm_reg_cvr_t is_on)
    def set_coverage(self, is_on):
        if is_on == UVM_NO_COVERAGE:
            self.m_cover_on = is_on
            return self.m_cover_on

        self.m_cover_on = self.m_has_cover & is_on
        return self.m_cover_on


    #   // Function: get_coverage
    #   //
    #   // Check if coverage measurement is on.
    #   //
    #   // Returns TRUE if measurement for all of the specified functional
    #   // coverage models are currently on.
    #   // Multiple functional coverage models can be specified by adding the
    #   // functional coverage model identifiers.
    #   //
    #   // See <uvm_mem::set_coverage()> for more details.
    #   //
    #   extern virtual function bit get_coverage(uvm_reg_cvr_t is_on)
    def get_coverage(self, is_on):
        if self.has_coverage(is_on) == 0:
            return 0
        return ((self.m_cover_on & is_on) == is_on)


    #   // Function: sample
    #   //
    #   // Functional coverage measurement method
    #   //
    #   // This method is invoked by the memory abstraction class
    #   // whenever an address within one of its address map
    #   // is successfully read or written.
    #   // The specified offset is the offset within the memory,
    #   // not an absolute address.
    #   //
    #   // Empty by default, this method may be extended by the
    #   // abstraction class generator to perform the required sampling
    #   // in any provided functional coverage model.
    #   //
    #   protected virtual function void  sample(uvm_reg_addr_t offset,
    #                                           bit            is_read,
    #                                           uvm_reg_map    map)
    def sample(self, offset, is_read, _map):
        pass


    def XsampleX(self, addr, is_read, _map):
        self.sample(addr, is_read, _map)


    #   // Core ovm_object operations
    #
    #   extern virtual function void do_print (uvm_printer printer)
    #   extern virtual function string convert2string()
    #   extern virtual function uvm_object clone()
    #   extern virtual function void do_copy   (uvm_object rhs)
    #   extern virtual function bit do_compare (uvm_object  rhs,
    #                                          uvm_comparer comparer)
    #   extern virtual function void do_pack (uvm_packer packer)
    #   extern virtual function void do_unpack (uvm_packer packer)
    #
    #
    #endclass: uvm_mem

#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#
#
#
#
#
#// set_offset
#
#function void uvm_mem::set_offset (uvm_reg_map    map,
#                                   uvm_reg_addr_t offset,
#                                   bit unmapped = 0)
#
#   uvm_reg_map orig_map = map
#
#   if (m_maps.num() > 1 && map is None):
#      `uvm_error("RegModel",{"set_offset requires a non-None map when memory '",
#                 get_full_name(),"' belongs to more than one map."})
#      return
#   end
#
#   map = get_local_map(map,"set_offset()")
#
#   if (map is None)
#     return
#
#   map.m_set_mem_offset(this, offset, unmapped)
#endfunction
#
#
#// is_in_map
#
#function bit uvm_mem::is_in_map(uvm_reg_map map)
#   if (m_maps.exists(map))
#     return 1
#   foreach (m_maps[l]):
#    uvm_reg_map local_map=l
#     uvm_reg_map parent_map = local_map.get_parent_map()
#
#     while (parent_map != None):
#       if (parent_map == map)
#         return 1
#       parent_map = parent_map.get_parent_map()
#     end
#   end
#   return 0
#endfunction
#
#
#// get_default_map
#
#function uvm_reg_map uvm_mem::get_default_map(string caller="")
#
#   // if mem is not associated with any may, return ~None~
#   if (m_maps.num() == 0):
#      `uvm_warning("RegModel",
#        {"Memory '",get_full_name(),"' is not registered with any map",
#         (caller == "" ? "": {" (called from ",caller,")"})})
#      return None
#   end
#
#   // if only one map, choose that
#   if (m_maps.num() == 1):
#     void'(m_maps.first(get_default_map))
#   end
#
#   // try to choose one based on default_map in parent blocks.
#   foreach (m_maps[l]):
#     uvm_reg_map map = l
#     uvm_reg_block blk = map.get_parent()
#     uvm_reg_map default_map = blk.get_default_map()
#     if (default_map != None):
#       uvm_reg_map local_map = get_local_map(default_map)
#       if (local_map != None)
#         return local_map
#     end
#   end
#
#   // if that fails, choose the first in this mem's maps
#
#   void'(m_maps.first(get_default_map))
#
#endfunction
#
#
#
#
#// get_offset
#
#function uvm_reg_addr_t uvm_mem::get_offset(uvm_reg_addr_t offset = 0,
#                                            uvm_reg_map map = None)
#
#   uvm_reg_map_info map_info
#   uvm_reg_map orig_map = map
#
#   map = get_local_map(map,"get_offset()")
#
#   if (map is None)
#     return -1
#
#   map_info = map.get_mem_map_info(this)
#
#   if (map_info.unmapped):
#      `uvm_warning("RegModel", {"Memory '",get_name(),
#                   "' is unmapped in map '",
#                   ((orig_map is None) ? map.get_full_name() : orig_map.get_full_name()),"'"})
#      return -1
#   end
#
#   return map_info.offset
#
#endfunction: get_offset
#
#
#
#// get_virtual_registers
#
#function void uvm_mem::get_virtual_registers(ref uvm_vreg regs[$])
#  foreach (m_vregs[vreg])
#     regs.append(vreg)
#endfunction
#
#
#// get_virtual_fields
#
#function void uvm_mem::get_virtual_fields(ref uvm_vreg_field fields[$])
#
#  foreach (m_vregs[l])
#  begin
#    uvm_vreg vreg = l
#    vreg.get_fields(fields)
#  end
#endfunction: get_virtual_fields
#
#
#// get_vfield_by_name
#
#function uvm_vreg_field uvm_mem::get_vfield_by_name(string name)
#  // Return first occurrence of vfield matching name
#  uvm_vreg_field vfields[$]
#
#  get_virtual_fields(vfields)
#
#  foreach (vfields[i])
#    if (vfields[i].get_name() == name)
#      return vfields[i]
#
#  `uvm_warning("RegModel", {"Unable to find virtual field '",name,
#                       "' in memory '",get_full_name(),"'"})
#   return None
#endfunction: get_vfield_by_name
#
#
#// get_vreg_by_name
#
#function uvm_vreg uvm_mem::get_vreg_by_name(string name)
#
#  foreach (m_vregs[l])
#  begin
#    uvm_vreg vreg = l
#    if (vreg.get_name() == name)
#      return vreg
#  end
#
#  `uvm_warning("RegModel", {"Unable to find virtual register '",name,
#                       "' in memory '",get_full_name(),"'"})
#  return None
#
#endfunction: get_vreg_by_name
#
#
#// get_vreg_by_offset
#
#function uvm_vreg uvm_mem::get_vreg_by_offset(uvm_reg_addr_t offset,
#                                              uvm_reg_map map = None)
#   `uvm_error("RegModel", "uvm_mem::get_vreg_by_offset() not yet implemented")
#   return None
#endfunction: get_vreg_by_offset
#
#
#
#// get_addresses
#
#function int uvm_mem::get_addresses(uvm_reg_addr_t offset = 0,
#                                    uvm_reg_map map=None,
#                                    ref uvm_reg_addr_t addr[])
#
#   uvm_reg_map_info map_info
#   uvm_reg_map system_map
#   uvm_reg_map orig_map = map
#
#   map = get_local_map(map,"get_addresses()")
#
#   if (map is None)
#     return 0
#
#   map_info = map.get_mem_map_info(this)
#
#   if (map_info.unmapped):
#      `uvm_warning("RegModel", {"Memory '",get_name(),
#                   "' is unmapped in map '",
#                   ((orig_map is None) ? map.get_full_name() : orig_map.get_full_name()),"'"})
#      return 0
#   end
#
#   addr = map_info.addr
#
#   foreach (addr[i])
#      addr[i] = addr[i] + map_info.mem_range.stride * offset
#
#   return map.get_n_bytes()
#
#endfunction
#
#
#// get_address
#
#function uvm_reg_addr_t uvm_mem::get_address(uvm_reg_addr_t offset = 0,
#                                             uvm_reg_map map = None)
#   uvm_reg_addr_t  addr[]
#   void'(get_addresses(offset, map, addr))
#   return addr[0]
#endfunction
#
#
#//-----------
#// HDL ACCESS
#//-----------
#
#
#
#
#
#
#//-------
#// ACCESS
#//-------
#
#
#
#
#
#//-----------------
#// Group- Frontdoor
#//-----------------
#
#// set_frontdoor
#
#function void uvm_mem::set_frontdoor(uvm_reg_frontdoor ftdr,
#                                     uvm_reg_map       map = None,
#                                     string            fname = "",
#                                     int               lineno = 0)
#   uvm_reg_map_info map_info
#   m_fname = fname
#   m_lineno = lineno
#
#   map = get_local_map(map, "set_frontdoor()")
#
#   if (map is None):
#      `uvm_error("RegModel", {"Memory '",get_full_name(),
#                 "' not found in map '", map.get_full_name(),"'"})
#      return
#   end
#
#   map_info = map.get_mem_map_info(this)
#   map_info.frontdoor = ftdr
#
#endfunction: set_frontdoor
#
#
#// get_frontdoor
#
#function uvm_reg_frontdoor uvm_mem::get_frontdoor(uvm_reg_map map = None)
#   uvm_reg_map_info map_info
#
#   map = get_local_map(map, "set_frontdoor()")
#
#   if (map is None):
#      `uvm_error("RegModel", {"Memory '",get_full_name(),
#                 "' not found in map '", map.get_full_name(),"'"})
#      return None
#   end
#
#   map_info = map.get_mem_map_info(this)
#   return map_info.frontdoor
#
#endfunction: get_frontdoor
#
#
#//----------------
#// Group- Backdoor
#//----------------
#
#
#
#
#
#// clear_hdl_path
#
#function void uvm_mem::clear_hdl_path(string kind = "RTL")
#  if (kind == "ALL"):
#    m_hdl_paths_pool = new("hdl_paths")
#    return
#  end
#
#  if (kind == "")
#    kind = m_parent.get_default_hdl_path()
#
#  if (!m_hdl_paths_pool.exists(kind)):
#    `uvm_warning("RegModel",{"Unknown HDL Abstraction '",kind,"'"})
#    return
#  end
#
#  m_hdl_paths_pool.delete(kind)
#endfunction
#
#
#// add_hdl_path
#
#function void uvm_mem::add_hdl_path(uvm_hdl_path_slice slices[], string kind = "RTL")
#    uvm_queue #(uvm_hdl_path_concat) paths = m_hdl_paths_pool.get(kind)
#    uvm_hdl_path_concat concat = new()
#
#    concat.set(slices)
#    paths.push_back(concat);
#endfunction
#
#
#
#
#
#
#// get_hdl_path
#
#function void uvm_mem::get_hdl_path(ref uvm_hdl_path_concat paths[$],
#                                    input string kind = "")
#
#  uvm_queue #(uvm_hdl_path_concat) hdl_paths
#
#  if (kind == "")
#     kind = m_parent.get_default_hdl_path()
#
#  if (!has_hdl_path(kind)):
#    `uvm_error("RegModel",
#        {"Memory does not have hdl path defined for abstraction '",kind,"'"})
#    return
#  end
#
#  hdl_paths = m_hdl_paths_pool.get(kind)
#
#  for (int i=0; i<hdl_paths.size();i++):
#     uvm_hdl_path_concat t = hdl_paths.get(i)
#     paths.push_back(t)
#  end
#
#endfunction
#
#
#// get_hdl_path_kinds
#
#function void uvm_mem::get_hdl_path_kinds (ref string kinds[$])
#  string kind
#  kinds.delete()
#  if (!m_hdl_paths_pool.first(kind))
#    return
#  do
#    kinds.push_back(kind)
#  while (m_hdl_paths_pool.next(kind))
#endfunction
#
#
#
#// set_parent
#
#function void uvm_mem::set_parent(uvm_reg_block parent)
#  m_parent = parent
#endfunction
#
#
#
#
#// convert2string
#
#function string uvm_mem::convert2string()
#
#   string res_str
#   string prefix
#
#   $sformat(convert2string, "%sMemory %s -- %0dx%0d bits", prefix,
#            get_full_name(), get_size(), get_n_bits())
#
#   if (m_maps.num()==0)
#     convert2string = {convert2string, "  (unmapped)\n"}
#   else
#     convert2string = {convert2string, "\n"}
#   foreach (m_maps[map]):
#     uvm_reg_map parent_map = map
#     int unsigned offset
#     while (parent_map != None):
#       uvm_reg_map this_map = parent_map
#       uvm_endianness_e endian_name
#       parent_map = this_map.get_parent_map()
#       endian_name=this_map.get_endian()
#
#       offset = parent_map is None ? this_map.get_base_addr(UVM_NO_HIER) :
#                                     parent_map.get_submap_offset(this_map)
#       prefix = {prefix, "  "}
#       $sformat(convert2string, "%sMapped in '%s' -- buswidth %0d bytes, %s, offset 'h%0h, size 'h%0h, %s\n", prefix,
#            this_map.get_full_name(), this_map.get_n_bytes(), endian_name.name(), offset,get_size(),get_access(this_map))
#     end
#   end
#   prefix = "  "
#   if (m_read_in_progress == 1'b1):
#      if (m_fname != "" && m_lineno != 0)
#         $sformat(res_str, "%s:%0d ",m_fname, m_lineno)
#      convert2string = {convert2string, "  ", res_str,
#                       "currently executing read method"};
#   end
#   if ( m_write_in_progress == 1'b1):
#      if (m_fname != "" && m_lineno != 0)
#         $sformat(res_str, "%s:%0d ",m_fname, m_lineno)
#      convert2string = {convert2string, "  ", res_str,
#                       "currently executing write method"};
#   end
#endfunction
#
#
#// do_print
#
#function void uvm_mem::do_print (uvm_printer printer)
#  super.do_print(printer)
#  //printer.print_generic(" ", " ", -1, convert2string())
#  printer.print_field_int("n_bits",get_n_bits(),32, UVM_UNSIGNED)
#  printer.print_field_int("size",get_size(),32, UVM_UNSIGNED)
#endfunction
#
#
#// clone
#
#function uvm_object uvm_mem::clone()
#  `uvm_fatal("RegModel","RegModel memories cannot be cloned")
#  return None
#endfunction
#
#// do_copy
#
#function void uvm_mem::do_copy(uvm_object rhs)
#  `uvm_fatal("RegModel","RegModel memories cannot be copied")
#endfunction
#
#
#// do_compare
#
#function bit uvm_mem::do_compare (uvm_object  rhs,
#                                        uvm_comparer comparer)
#  `uvm_warning("RegModel","RegModel memories cannot be compared")
#  return 0
#endfunction
#
#
#// do_pack
#
#function void uvm_mem::do_pack (uvm_packer packer)
#  `uvm_warning("RegModel","RegModel memories cannot be packed")
#endfunction
#
#
#// do_unpack
#
#function void uvm_mem::do_unpack (uvm_packer packer)
#  `uvm_warning("RegModel","RegModel memories cannot be unpacked")
#endfunction
#
#
#
#
#// Xdelete_vregX
#
#function void uvm_mem::Xdelete_vregX(uvm_vreg vreg)
#   if (m_vregs.exists(vreg))
#     m_vregs.delete(vreg)
#endfunction
#
#
